#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
This module provides an interface to elasticsearch. Other parts of NOMAD must not
interact with elasticsearch to maintain a clear coherent interface and allow for change.

Currently NOMAD uses one entry index and two distinct materials indices. The entries
index is based on two different mappings, once used by the old flask api (v0) and one
used by the new fastapi api (v1). The mappings are used at the same time and the documents
are merged. Write operations (index, publish, edit, lift embargo, delete) are common; defined
here in the module ``__init__.py``. Read operations are different and
should be used as per use-case directly from the ``v0`` and ``v1`` submodules.

Most common functions also take an ``update_materials`` keyword arg with allows to
update the v1 materials index according to the performed changes. TODO this is only
partially implemented.
'''

from typing import Union, List, Iterable, Any, cast, Dict, Iterator, Generator, Callable
import json
import elasticsearch.helpers
from elasticsearch.exceptions import TransportError, RequestError
from elasticsearch_dsl import Q, A, Search
from elasticsearch_dsl.query import Query as EsQuery
from pydantic.error_wrappers import ErrorWrapper
from pydantic import ValidationError

from nomad import config, infrastructure, utils
from nomad import datamodel
from nomad.app.v1 import models
from nomad.datamodel import EntryArchive, EntryMetadata, user_reference, author_reference
from nomad.app.v1.models import (
    AggregationPagination, Criteria, MetadataPagination, Pagination, PaginationResponse,
    QuantityAggregation, Query, MetadataRequired,
    MetadataResponse, Aggregation, StatisticsAggregation, StatisticsAggregationResponse,
    Value, AggregationBase, TermsAggregation, BucketAggregation, HistogramAggregation,
    DateHistogramAggregation, MinMaxAggregation, Bucket,
    MinMaxAggregationResponse, TermsAggregationResponse, HistogramAggregationResponse,
    DateHistogramAggregationResponse, AggregationResponse)
from nomad.metainfo.elasticsearch_extension import (
    index_entries, entry_type, entry_index, DocumentType,
    material_type, entry_type, material_entry_type,
    entry_index, Index, index_entries, DocumentType, SearchQuantity, update_materials)


def update_by_query(
        update_script: str,
        query: Any = None,
        owner: str = None,
        user_id: str = None,
        index: str = None,
        refresh: bool = False,
        **kwargs):
    '''
    Uses the given painless script to update the entries by given query.

    In most cases, the elasticsearch entry index should not be updated field by field;
    you should run `index` instead and fully replace documents from mongodb and
    archive files.

    This method provides a faster direct method to update individual fields, e.g. to quickly
    update fields for editing operations.
    '''
    if query is None:
        query = {}

    es_query_normalized = normalize_api_query(cast(Query, query), doc_type=entry_type)
    owner_query = _owner_es_query(owner=owner, user_id=user_id, doc_type=entry_type)
    es_query_validated = validate_api_query(es_query_normalized, entry_type, owner_query)

    body = {
        'script': {
            'source': update_script,
            'lang': 'painless'
        },
        'query': es_query_validated.to_dict()
    }

    body['script'].update(**kwargs)

    try:
        result = infrastructure.elastic_client.update_by_query(
            body=body, index=config.elastic.entries_index)
    except TransportError as e:
        utils.get_logger(__name__).error(
            'es update_by_query script error', exc_info=e,
            es_info=json.dumps(e.info, indent=2))
        raise SearchError(e)

    if refresh:
        _refresh()

    return result


def delete_by_query(
        query: dict,
        owner: str = None,
        user_id: str = None,
        update_materials: bool = False,
        refresh: bool = False):
    '''
    Deletes all entries that match the given query.
    '''
    if query is None:
        query = {}

    es_query_normalized = normalize_api_query(cast(Query, query), doc_type=entry_type)
    owner_query = _owner_es_query(owner=owner, user_id=user_id, doc_type=entry_type)
    es_query_validated = validate_api_query(es_query_normalized, entry_type, owner_query)

    body = {
        'query': es_query_validated.to_dict()
    }

    try:
        result = infrastructure.elastic_client.delete_by_query(
            body=body, index=config.elastic.entries_index)
    except TransportError as e:
        utils.get_logger(__name__).error(
            'es delete_by_query error', exc_info=e,
            es_info=json.dumps(e.info, indent=2))
        raise SearchError(e)

    if refresh:
        _refresh()

    if update_materials:
        # TODO update the matrials index at least for v1
        pass

    return result


def refresh():
    '''
    Refreshes the specified indices.
    '''

    try:
        infrastructure.elastic_client.indices.refresh(index=config.elastic.entries_index)
    except TransportError as e:
        utils.get_logger(__name__).error(
            'es delete_by_query error', exc_info=e,
            es_info=json.dumps(e.info, indent=2))
        raise SearchError(e)


_refresh = refresh


def index(
        entries: Union[EntryArchive, List[EntryArchive]], update_materials: bool = False,
        refresh: bool = False):
    '''
    Index the given entries based on their archive. Either creates or updates the underlying
    elasticsearch documents. If an underlying elasticsearch document already exists it
    will be fully replaced.
    '''
    if not isinstance(entries, list):
        entries = [entries]

    index_entries(entries, refresh=refresh or update_materials)
    if update_materials:
        index_materials(entries, refresh=refresh)


def index_materials(entries: Union[EntryArchive, List[EntryArchive]], **kwargs):
    '''
    Index the materials within the given entries based on their archive. The entries
    have to be indexed first.
    '''

    if not isinstance(entries, list):
        entries = [entries]

    update_materials(entries=entries, **kwargs)


# TODO this depends on how we merge section metadata
def publish(entries: Iterable[EntryMetadata], index: str = None) -> int:
    '''
    Publishes the given entries based on their entry metadata. Sets publishes to true,
    and updates most user provided metadata with a partial update. Returns the number
    of failed updates.
    '''
    return update_metadata(
        entries, index=index, published=True, update_materials=True, refresh=True)


def update_metadata(
        entries: Iterable[EntryMetadata], index: str = None,
        update_materials: bool = False, refresh: bool = False,
        **kwargs) -> int:
    '''
    Update all given entries with their given metadata. Additionally apply kwargs.
    Returns the number of failed updates. This is doing a partial update on the underlying
    elasticsearch documents.
    '''

    def elastic_updates():
        for entry_metadata in entries:
            entry_archive = entry_metadata.m_parent
            if entry_archive is None:
                entry_archive = EntryArchive(metadata=entry_metadata)
            entry_doc = entry_type.create_index_doc(entry_archive)

            entry_doc.update(**kwargs)

            yield dict(
                doc=entry_doc,
                _id=entry_metadata.entry_id,
                _index=entry_index.index_name,
                _op_type='update')

    updates = list(elastic_updates())
    _, failed = elasticsearch.helpers.bulk(
        infrastructure.elastic_client, updates, stats_only=True)
    failed = cast(int, failed)

    if update_materials:
        # TODO update the matrials index at least for v1
        pass

    if refresh:
        _refresh()

    return failed


def delete_upload(upload_id: str, refresh: bool = False, **kwargs):
    '''
    Deletes the given upload.
    '''
    delete_by_query(query=dict(upload_id=upload_id), **kwargs)

    if refresh:
        _refresh()


def delete_entry(entry_id: str, index: str = None, refresh: bool = False, **kwargs):
    '''
    Deletes the given entry.
    '''
    delete_by_query(query=dict(entry_id=entry_id), **kwargs)

    if refresh:
        _refresh()


class SearchError(Exception): pass


class AuthenticationRequiredError(Exception): pass


_entry_metadata_defaults = {
    quantity.name: quantity.default
    for quantity in datamodel.EntryMetadata.m_def.quantities  # pylint: disable=not-an-iterable
    if quantity.default not in [None, [], False, 0]
}

_all_author_quantities = [
    quantity.name
    for quantity in EntryMetadata.m_def.all_quantities.values()
    if quantity.type in [user_reference, author_reference]]


def _es_to_entry_dict(hit, required: MetadataRequired = None) -> Dict[str, Any]:
    '''
    Elasticsearch entry metadata does not contain default values, if a metadata is not
    set. This will add default values to entry metadata in dict form obtained from
    elasticsearch.
    '''
    entry_dict = hit.to_dict()
    for key, value in _entry_metadata_defaults.items():
        if key not in entry_dict:
            if required is not None:
                if required.exclude and key in required.exclude:
                    continue
                if required.include and key not in required.include:
                    continue

            entry_dict[key] = value

    for author_quantity in _all_author_quantities:
        authors = entry_dict.get(author_quantity)
        if authors is None:
            continue
        if isinstance(authors, dict):
            authors = [authors]
        for author in authors:
            if 'email' in author:
                del(author['email'])

    return entry_dict


def _owner_es_query(owner: str, user_id: str = None, doc_type: DocumentType = entry_type):
    def term_query(**kwargs):
        prefix = '' if doc_type == entry_type else 'entries.'
        return Q('term', **{
            (prefix + field): value for field, value in kwargs.items()})

    if owner == 'all':
        q = term_query(published=True)
        if user_id is not None:
            q = q | term_query(viewers__user_id=user_id)
    elif owner == 'public':
        q = term_query(published=True) & term_query(with_embargo=False)
    elif owner == 'visible':
        q = term_query(published=True) & term_query(with_embargo=False)
        if user_id is not None:
            q = q | term_query(viewers__user_id=user_id)
    elif owner == 'shared':
        if user_id is None:
            raise AuthenticationRequiredError('Authentication required for owner value shared.')

        q = term_query(viewers__user_id=user_id)
    elif owner == 'user':
        if user_id is None:
            raise AuthenticationRequiredError('Authentication required for owner value user.')

        q = term_query(main_author__user_id=user_id)
    elif owner == 'staging':
        if user_id is None:
            raise AuthenticationRequiredError('Authentication required for owner value user')
        q = term_query(published=False) & term_query(viewers__user_id=user_id)
    elif owner == 'admin':
        if user_id is None or not datamodel.User.get(user_id=user_id).is_admin:
            raise AuthenticationRequiredError('This can only be used by the admin user.')
        q = None
    elif owner is None:
        q = None
    else:
        raise KeyError('Unsupported owner value')

    if q is not None:
        return q
    return Q()


class QueryValidationError(Exception):
    def __init__(self, error, loc):
        self.errors = [ErrorWrapper(Exception(error), loc=loc)]


def validate_quantity(
        quantity_name: str, value: Value = None, doc_type: DocumentType = None,
        loc: List[str] = None) -> SearchQuantity:
    '''
    Validates the given quantity name and value against the given document type.

    Returns:
        A metainfo elasticsearch extension SearchQuantity object.

    Raises: QueryValidationError
    '''
    assert quantity_name is not None

    if doc_type == material_entry_type and not quantity_name.startswith('entries'):
        quantity_name = f'entries.{quantity_name}'

    if doc_type == material_type and quantity_name.startswith('entries'):
        doc_type = material_entry_type

    if doc_type is None:
        doc_type = entry_type

    quantity = doc_type.quantities.get(quantity_name)
    if quantity is None:
        raise QueryValidationError(
            f'{quantity_name} is not a {doc_type} quantity',
            loc=[quantity_name] if loc is None else loc)

    return quantity


def normalize_api_query(query: Query, doc_type: DocumentType, prefix: str = None) -> Query:
    '''
    Normalizes the given query. Should be applied before validate_api_query, which
    expects a normalized query. Normalization will
    - replace nested dicts with`models.And`, `models.Nested` instances
    - introduce `models.Nested` if necessary
    - replace dicts with `models.And` or 'models.Range' queries.

    After normalization there should be no dicts or `*:(any|all|none)` values in the query.
    '''
    def normalize_criteria(name, value: models.CriteriaValue, prefix: str) -> Query:
        if prefix is not None:
            full_name = f'{prefix}.{name}'
        else:
            full_name = name

        prefixes = []
        name_wo_prefix = name
        nested_prefix = None
        for nested_key in doc_type.nested_object_keys:
            if nested_key == prefix:
                continue

            if full_name.startswith(f'{nested_key}'):
                if prefix is None or not prefix.startswith(nested_key):
                    prefixes.append(nested_key)
                    name_wo_prefix = full_name[len(nested_key) + 1:]
                    nested_prefix = nested_key

            if full_name == nested_key:
                break
        name = name_wo_prefix

        query: Query = None

        # Dictionaries that are not at the root level can either be range
        # queries, or else they are interpreted as a list of AND queries.
        if isinstance(value, dict):
            try:
                query = Criteria(name=name, value=models.Range(**value))
            except ValidationError:
                query = models.And(**{'and': [
                    normalize_criteria(k if name == '' else f'{name}.{k}', v, nested_prefix)
                    for k, v in value.items()]})

        else:
            query = Criteria(name=name, value=value)

        for prefix in reversed(prefixes):
            query = models.Nested(prefix=prefix, query=query)

        return query

    def normalize_query(query: Query):
        return normalize_api_query(query, doc_type=doc_type, prefix=prefix)

    if isinstance(query, dict):
        if len(query) is None:
            return models.Empty()

        if len(query) == 1:
            name = next(iter(query))
            return normalize_criteria(name, query[name], prefix)

        return models.And(**{'and': [
            normalize_criteria(name, value, prefix) for name, value in query.items()]})

    if isinstance(query, models.And):
        return models.And(**{'and': [normalize_query(op) for op in query.op]})

    if isinstance(query, models.Or):
        return models.Or(**{'or': [normalize_query(op) for op in query.op]})

    if isinstance(query, models.Not):
        return models.Not(**{'not': normalize_query(query.op)})

    if isinstance(query, models.Nested):
        return models.Nested(
            prefix=query.prefix,
            query=normalize_api_query(query, doc_type=doc_type, prefix=query.prefix))

    if isinstance(query, (models.Empty, models.Criteria)):
        return query

    raise NotImplementedError(f'Query type {query.__class__} is not supported')


def remove_quantity_from_query(query: Query, quantity: str, prefix=None):
    '''
    Removes all criteria with the given quantity from the query. Query has to be
    normalized. Remove is done by replacing respective criteria with an empty query.
    '''

    if isinstance(query, models.And):
        return models.And(**{'and': [remove_quantity_from_query(op, quantity, prefix) for op in query.op]})

    if isinstance(query, models.Or):
        return models.Or(**{'or': [remove_quantity_from_query(op, quantity, prefix) for op in query.op]})

    if isinstance(query, models.Not):
        return models.Not(**{'not': remove_quantity_from_query(query.op, quantity, prefix)})

    if isinstance(query, models.Nested):
        return models.Nested(
            prefix=query.prefix,
            query=remove_quantity_from_query(query.query, quantity, prefix=query.prefix))

    if isinstance(query, models.Empty):
        return query

    if isinstance(query, models.Criteria):
        name = query.name
        if prefix is not None:
            name = f'{prefix}.{name}'
        if name == quantity:
            return models.Empty()

        return query

    raise NotImplementedError(f'Query type {query.__class__} is not supported')


def validate_api_query(
        query: Query, doc_type: DocumentType, owner_query: EsQuery, prefix: str = None) -> EsQuery:
    '''
    Creates an ES query based on the API's query model. This needs to be a normalized
    query.

    However, this function performs validation of quantities and types and raises
    a QueryValidationError accordingly. This exception is populated with pydantic
    errors.

    Arguments:
        query: The api query object.
        doc_type:
            The elasticsearch metainfo extension document type that this query needs to
            be verified against.
        owner_query:
            A prebuild ES query that is added to nested entries query. Only for
            materials queries.
        prefix:
            An optional prefix that is added to all quantity names. Used for recursion.

    Returns:
        A elasticsearch dsl query object.

    Raises: QueryValidationError
    '''

    def match(name: str, value: Value) -> EsQuery:
        if prefix is not None:
            name = f'{prefix}.{name}'

        if name == 'optimade_filter':
            value = str(value)
            from nomad.app.optimade import filterparser
            try:
                return filterparser.parse_filter(value, without_prefix=True)

            except filterparser.FilterException as e:
                raise QueryValidationError(
                    f'Could not parse optimade filter: {e}',
                    loc=[name])

        # TODO non keyword quantities, type checks
        quantity = validate_quantity(name, value, doc_type=doc_type)
        normalizer = quantity.annotation.normalizer
        if normalizer:
            value = normalizer(value)
        return Q(quantity.annotation.es_query, **{quantity.search_field: value})

    def validate_query(query: Query) -> EsQuery:
        return validate_api_query(
            query, doc_type=doc_type, owner_query=owner_query, prefix=prefix)

    def validate_criteria(name: str, value: Any):
        if isinstance(value, models.All):
            return Q('bool', must=[match(name, item) for item in value.op])

        elif isinstance(value, models.Any_):
            return Q('bool', should=[match(name, item) for item in value.op])

        elif isinstance(value, models.None_):
            return Q('bool', must_not=[match(name, item) for item in value.op])

        elif isinstance(value, models.Range):
            if prefix is not None:
                name = f'{prefix}.{name}'
            quantity = validate_quantity(name, None, doc_type=doc_type)
            return Q('range', **{quantity.search_field: value.dict(
                exclude_unset=True,
            )})

        elif isinstance(value, (models.And, models.Or, models.Not)):
            return validate_query(value)

        # list of values is treated as an "all" over the items
        elif isinstance(value, list):
            return Q('bool', must=[match(name, item) for item in value])

        elif isinstance(value, dict):
            raise NotImplementedError()

        else:
            return match(name, value)

    if isinstance(query, models.And):
        return Q('bool', must=[validate_query(operand) for operand in query.op])

    if isinstance(query, models.Or):
        return Q('bool', should=[validate_query(operand) for operand in query.op])

    if isinstance(query, models.Not):
        return Q('bool', must_not=validate_query(query.op))

    if isinstance(query, models.Nested):
        sub_doc_type = material_entry_type if query.prefix == 'entries' else doc_type
        sub_query = validate_api_query(
            query.query, doc_type=sub_doc_type, prefix=query.prefix, owner_query=owner_query)

        if query.prefix == 'entries':
            sub_query &= owner_query

        return Q('nested', path=query.prefix, query=sub_query)

    if isinstance(query, models.Criteria):
        return validate_criteria(query.name, query.value)

    if isinstance(query, models.Empty):
        return Q()

    raise NotImplementedError(f'Query type {query.__class__} is not supported')


def validate_pagination(pagination: Pagination, doc_type: DocumentType, loc: List[str] = None):
    order_quantity = None
    if pagination.order_by is not None:
        order_quantity = validate_quantity(
            pagination.order_by, doc_type=doc_type, loc=['pagination', 'order_by'])
        if not order_quantity.definition.is_scalar:
            raise QueryValidationError(
                'the order_by quantity must be a scalar',
                loc=(loc if loc else []) + ['pagination', 'order_by'])

    page_after_value = pagination.page_after_value
    if page_after_value is not None and \
            pagination.order_by is not None and \
            pagination.order_by != doc_type.id_field and \
            ':' not in page_after_value:

        pagination.page_after_value = '%s:' % page_after_value

    return order_quantity, page_after_value


def _api_to_es_aggregation(
        es_search: Search, name: str, agg: AggregationBase, doc_type: DocumentType,
        post_agg_query: models.Query, create_es_query: Callable[[models.Query], EsQuery]) -> A:
    '''
    Creates an ES aggregation based on the API's aggregation model.
    '''

    agg_name = f'agg:{name}'
    es_aggs = es_search.aggs

    if post_agg_query:
        if isinstance(agg, QuantityAggregation) and agg.exclude_from_search:
            filter = create_es_query(remove_quantity_from_query(post_agg_query, agg.quantity))
        else:
            filter = create_es_query(post_agg_query)
        es_aggs = es_aggs.bucket(f'{agg_name}:filtered', A('filter', filter=filter))

    if isinstance(agg, StatisticsAggregation):
        for metric_name in agg.metrics:
            metrics = doc_type.metrics
            if metric_name not in metrics and doc_type == material_type:
                metrics = material_entry_type.metrics
            if metric_name not in metrics:
                raise QueryValidationError(
                    'metric must be the qualified name of a suitable search quantity',
                    loc=['statistic', 'metrics'])
            metric_aggregation, metric_quantity = metrics[metric_name]
            es_aggs.metric('statistics:%s' % metric_name, A(
                metric_aggregation,
                field=metric_quantity.qualified_field))

        return

    agg = cast(QuantityAggregation, agg)
    longest_nested_key = None
    is_nested = False
    quantity = validate_quantity(agg.quantity, doc_type=doc_type, loc=['aggregation', 'quantity'])
    for nested_key in doc_type.nested_object_keys:
        if agg.quantity.startswith(nested_key):
            es_aggs = es_aggs.bucket('nested_agg:%s' % name, 'nested', path=nested_key)
            longest_nested_key = nested_key
            is_nested = True

    es_agg = None

    if isinstance(agg, TermsAggregation):
        if not quantity.aggregateable:
            raise QueryValidationError(
                'The aggregation quantity cannot be used in a terms aggregation.',
                loc=['aggregation', name, 'terms', 'quantity'])

        if agg.pagination is not None:
            if post_agg_query is not None:
                raise QueryValidationError(
                    f'aggregation pagination cannot be used with exclude_from_search in the same request',
                    loc=['aggregations', name, 'terms', 'pagination'])

            if agg.size is not None:
                raise QueryValidationError(
                    f'You cannot paginate and provide an extra size parameter.',
                    loc=['aggregations', name, 'terms', 'pagination'])

            order_quantity, page_after_value = validate_pagination(
                agg.pagination, doc_type=doc_type, loc=['aggregation'])

            # We are using elastic searchs 'composite aggregations' here. We do not really
            # compose aggregations, but only those pseudo composites allow us to use the
            # 'after' feature that allows to scan through all aggregation values.
            terms = A('terms', field=quantity.search_field, order=agg.pagination.order.value)

            if order_quantity is None:
                composite = {
                    'sources': {
                        name: terms
                    },
                    'size': agg.pagination.page_size
                }

            else:
                sort_terms = A(
                    'terms',
                    field=order_quantity.search_field,
                    order=agg.pagination.order.value)

                composite = {
                    'sources': [
                        {order_quantity.search_field: sort_terms},
                        {quantity.search_field: terms}
                    ],
                    'size': agg.pagination.page_size
                }

            if page_after_value is not None:
                if order_quantity is None:
                    composite['after'] = {name: page_after_value}
                else:
                    try:
                        order_value, quantity_value = page_after_value.split(':')
                        composite['after'] = {quantity.search_field: quantity_value, order_quantity.search_field: order_value}
                    except Exception:
                        raise QueryValidationError(
                            f'The pager_after_value has not the right format.',
                            loc=['aggregations', name, 'terms', 'pagination', 'page_after_value'])

            es_agg = es_aggs.bucket(agg_name, 'composite', **composite)

            # additional cardinality to get total
            es_aggs.metric('agg:%s:total' % name, 'cardinality', field=quantity.search_field)
        else:
            if agg.size is None:
                if quantity.default_aggregation_size is not None:
                    agg.size = quantity.default_aggregation_size

                elif quantity.values is not None:
                    agg.size = len(quantity.values)

                else:
                    agg.size = 10

            terms_kwargs: Dict[str, Any] = {}
            if agg.include is not None:
                if isinstance(agg.include, str):
                    terms_kwargs["include"] = f'.*{agg.include}.*'
                else:
                    terms_kwargs["include"] = agg.include

            terms = A('terms', field=quantity.search_field, size=agg.size, **terms_kwargs)
            es_agg = es_aggs.bucket(agg_name, terms)

        if agg.entries is not None and agg.entries.size > 0:
            kwargs: Dict[str, Any] = {}
            if agg.entries.required is not None:
                if agg.entries.required.include is not None:
                    kwargs.update(_source=dict(includes=agg.entries.required.include))
                else:
                    kwargs.update(_source=dict(excludes=agg.entries.required.exclude))

            es_agg.metric('entries', A('top_hits', size=agg.entries.size, **kwargs))

        if is_nested:
            es_agg.bucket(f'agg:parents:{name}', A('reverse_nested'))

    elif isinstance(agg, DateHistogramAggregation):
        if not quantity.annotation.mapping['type'] in ['date']:
            raise QueryValidationError(
                f'The quantity {quantity} cannot be used in a date histogram aggregation',
                loc=['aggregations', name, 'histogram', 'quantity'])

        es_agg = es_aggs.bucket(agg_name, A(
            'date_histogram', field=quantity.search_field, interval=agg.interval,
            format='yyyy-MM-dd'))

    elif isinstance(agg, HistogramAggregation):
        if not quantity.annotation.mapping['type'] in ['integer', 'float', 'double', 'long']:
            raise QueryValidationError(
                f'The quantity {quantity} cannot be used in a histogram aggregation',
                loc=['aggregations', name, 'histogram', 'quantity'])

        es_agg = es_aggs.bucket(agg_name, A(
            'histogram', field=quantity.search_field, interval=agg.interval))

    elif isinstance(agg, MinMaxAggregation):
        if not quantity.annotation.mapping['type'] in ['integer', 'float', 'double', 'long', 'date']:
            raise QueryValidationError(
                f'The quantity {quantity} cannot be used in a mix-max aggregation',
                loc=['aggregations', name, 'min_max', 'quantity'])

        es_aggs.metric(agg_name + ':min', A('min', field=quantity.search_field))
        es_aggs.metric(agg_name + ':max', A('max', field=quantity.search_field))

    else:
        raise NotImplementedError()

    if isinstance(agg, BucketAggregation):
        for metric_name in agg.metrics:
            metrics = doc_type.metrics
            if longest_nested_key == 'entries':
                metrics = material_entry_type.metrics
            if metric_name not in metrics:
                raise QueryValidationError(
                    'metric must be the qualified name of a suitable search quantity',
                    loc=['statistic', 'metrics'])
            metric_aggregation, metric_quantity = metrics[metric_name]
            es_agg.metric('metric:%s' % metric_name, A(
                metric_aggregation,
                field=metric_quantity.qualified_field))


def _es_to_api_aggregation(
        es_response, name: str, agg: AggregationBase, doc_type: DocumentType):
    '''
    Creates a AggregationResponse from elasticsearch response on a request executed with
    the given aggregation.
    '''
    es_aggs = es_response.aggs

    filtered_agg_name = f'agg:{name}:filtered'
    if filtered_agg_name in es_response.aggs:
        es_aggs = es_aggs[f'agg:{name}:filtered']

    aggregation_dict = agg.dict(by_alias=True)

    if isinstance(agg, StatisticsAggregation):
        metrics = {}
        for metric in agg.metrics:  # type: ignore
            metrics[metric] = es_aggs[f'statistics:{metric}'].value

        return AggregationResponse(
            statistics=StatisticsAggregationResponse(data=metrics, **aggregation_dict))

    agg = cast(QuantityAggregation, agg)
    quantity = validate_quantity(agg.quantity, doc_type=doc_type)
    longest_nested_key = None
    for nested_key in doc_type.nested_object_keys:
        if agg.quantity.startswith(nested_key):
            es_aggs = es_aggs[f'nested_agg:{name}']
            longest_nested_key = nested_key

    has_no_pagination = getattr(agg, 'pagination', None) is None

    if isinstance(agg, BucketAggregation):
        es_agg = es_aggs['agg:' + name]
        values = set()

        def get_bucket(es_bucket) -> Bucket:
            if has_no_pagination:
                if isinstance(agg, DateHistogramAggregation):
                    value = es_bucket['key_as_string']
                else:
                    value = es_bucket['key']
            elif agg.pagination.order_by is None:  # type: ignore
                value = es_bucket.key[name]
            else:
                value = es_bucket.key[quantity.search_field]

            nested_count = es_bucket.doc_count
            if f'agg:parents:{name}' in es_bucket:
                count = es_bucket[f'agg:parents:{name}'].doc_count
            else:
                count = nested_count
            metrics = {}
            for metric in agg.metrics:  # type: ignore
                metrics[metric] = es_bucket['metric:' + metric].value

            entries = None
            if 'entries' in es_bucket:
                if longest_nested_key:
                    entries = [{longest_nested_key: item['_source'].to_dict()} for item in es_bucket.entries.hits.hits]
                else:
                    entries = [item['_source'].to_dict() for item in es_bucket.entries.hits.hits]

            # By default ES returns values of 0 and 1 for terms aggregation
            # targeting boolean values. Here we transform them into True/False
            # to be more consistent.
            if isinstance(agg, TermsAggregation) and quantity.annotation.mapping["type"] == "boolean":
                if value == 0:
                    value = False
                elif value == 1:
                    value = True

            values.add(value)
            if len(metrics) == 0:
                metrics = None
            return Bucket(
                value=value, entries=entries, count=count, nested_count=nested_count,
                metrics=metrics)

        data = [get_bucket(es_bucket) for es_bucket in es_agg.buckets]

        if has_no_pagination:
            # fill "empty" values
            if quantity.values is not None:
                for value in quantity.values:
                    if value not in values:
                        metrics = {metric: 0 for metric in agg.metrics}
                        if len(metrics) == 0:
                            metrics = None
                        data.append(Bucket(value=value, count=0, metrics=metrics))

        else:
            total = es_aggs['agg:%s:total' % name]['value']
            pagination = PaginationResponse(total=total, **aggregation_dict['pagination'])
            if pagination.page_after_value is not None and pagination.page_after_value.endswith(':'):
                pagination.page_after_value = pagination.page_after_value[0:-1]

            if 'after_key' in es_agg:
                after_key = es_agg['after_key']
                if pagination.order_by is None:
                    pagination.next_page_after_value = after_key[name]
                else:
                    str_values = [str(v) for v in after_key.to_dict().values()]
                    pagination.next_page_after_value = ':'.join(str_values)
            else:
                pagination.next_page_after_value = None

            aggregation_dict['pagination'] = pagination

        if isinstance(agg, TermsAggregation):
            return AggregationResponse(
                terms=TermsAggregationResponse(data=data, **aggregation_dict))
        elif isinstance(agg, HistogramAggregation):
            return AggregationResponse(
                histogram=HistogramAggregationResponse(data=data, **aggregation_dict))
        elif isinstance(agg, DateHistogramAggregation):
            return AggregationResponse(
                date_histogram=DateHistogramAggregationResponse(data=data, **aggregation_dict))
        else:
            raise NotImplementedError()

    if isinstance(agg, MinMaxAggregation):
        min_value = es_aggs['agg:%s:min' % name]['value']
        max_value = es_aggs['agg:%s:max' % name]['value']

        return AggregationResponse(
            min_max=MinMaxAggregationResponse(data=[min_value, max_value], **aggregation_dict))

    raise NotImplementedError()


def _specific_agg(agg: Aggregation) -> Union[TermsAggregation, DateHistogramAggregation, HistogramAggregation, MinMaxAggregation, StatisticsAggregation]:
    if agg.terms is not None:
        return agg.terms

    if agg.histogram is not None:
        return agg.histogram

    if agg.date_histogram is not None:
        return agg.date_histogram

    if agg.min_max is not None:
        return agg.min_max

    if agg.statistics is not None:
        return agg.statistics

    raise NotImplementedError()


def _and_clauses(query: Query) -> Generator[Query, None, None]:
    if isinstance(query, models.And):
        for clause in query.op:
            for query in _and_clauses(clause):
                yield query

    yield query


def search(
        owner: str = 'public',
        query: Union[Query, EsQuery] = None,
        pagination: MetadataPagination = None,
        required: MetadataRequired = None,
        aggregations: Dict[str, Aggregation] = {},
        user_id: str = None,
        index: Index = entry_index) -> MetadataResponse:

    # The first half of this method creates the ES query. Then the query is run on ES.
    # The second half is about transforming the ES response to a MetadataResponse.

    doc_type = index.doc_type

    # owner
    owner_query = _owner_es_query(owner=owner, user_id=user_id, doc_type=doc_type)

    # query
    if query is None:
        query = {}

    def create_es_query(query: Query):
        return validate_api_query(cast(Query, query), doc_type=doc_type, owner_query=owner_query)

    if isinstance(query, EsQuery):
        es_query = cast(EsQuery, query)
    else:
        query = normalize_api_query(cast(Query, query), doc_type=doc_type)
        es_query = create_es_query(cast(Query, query))

    nested_owner_query = owner_query
    if doc_type != entry_type:
        nested_owner_query = Q('nested', path='entries', query=owner_query)
    es_query &= nested_owner_query

    # pagination
    if pagination is None:
        pagination = MetadataPagination()

    if pagination.order_by is None:
        pagination.order_by = doc_type.id_field

    search = Search(index=index.index_name)

    # TODO this depends on doc_type
    if pagination.order_by is None:
        pagination.order_by = doc_type.id_field
    order_quantity, page_after_value = validate_pagination(pagination, doc_type=doc_type)
    order_field = order_quantity.search_field
    sort = {order_field: pagination.order.value}
    if order_field != doc_type.id_field:
        sort[doc_type.id_field] = pagination.order.value
    search = search.sort(sort)
    search = search.extra(size=pagination.page_size, track_total_hits=True)

    if pagination.page_offset:
        search = search.extra(**{'from': pagination.page_offset})
    elif pagination.page:
        search = search.extra(**{'from': (pagination.page - 1) * pagination.page_size})
    elif page_after_value:
        search = search.extra(search_after=page_after_value.rsplit(':', 1))

    # required
    excludes = ["*__suggestion"]  # Suggestion values are always excluded
    includes = None
    if required:
        for list_ in [required.include, required.exclude]:
            for quantity in [] if list_ is None else list_:
                # TODO validate quantities with wildcards
                if '*' not in quantity:
                    validate_quantity(quantity, doc_type=doc_type, loc=['required'])

        if required.include is not None and pagination.order_by not in required.include:
            required.include.append(pagination.order_by)
        if required.exclude is not None and pagination.order_by in required.exclude:
            required.exclude.remove(pagination.order_by)

        if required.include is not None and doc_type.id_field not in required.include:
            required.include.append(doc_type.id_field)

        if required.exclude is not None and doc_type.id_field in required.exclude:
            required.exclude.remove(doc_type.id_field)

        if required.exclude:
            excludes += required.exclude
        includes = required.include
    search = search.source(includes=includes, excludes=excludes)

    # aggregations
    aggs = [(name, _specific_agg(agg)) for name, agg in aggregations.items()]
    excluded_agg_quantities = {
        agg.quantity
        for _, agg in aggs
        if isinstance(agg, QuantityAggregation) and agg.exclude_from_search}

    if len(excluded_agg_quantities) > 0:
        and_clauses = list(_and_clauses(query))
        pre_clauses = [
            and_clause for and_clause in and_clauses
            if isinstance(and_clause, models.Criteria) and and_clause.name not in excluded_agg_quantities]

        pre_agg_es_query = validate_api_query(
            models.And(**{'and': list(pre_clauses)}), doc_type=doc_type,
            owner_query=owner_query)
        post_agg_query = models.And(**{'and': [
            and_clause for and_clause in and_clauses if and_clause not in pre_clauses]})
        post_agg_es_query = validate_api_query(
            post_agg_query, doc_type=doc_type, owner_query=owner_query)

        search = search.post_filter(post_agg_es_query)
        search = search.query(pre_agg_es_query & nested_owner_query)

    else:
        search = search.query(es_query)  # pylint: disable=no-member
        post_agg_query = None

    for name, agg in aggs:
        _api_to_es_aggregation(
            search, name, agg, doc_type=doc_type,
            post_agg_query=post_agg_query, create_es_query=create_es_query)

    # execute
    try:
        es_response = search.execute()
    except RequestError as e:
        raise SearchError(e)
    more_response_data = {}

    # pagination
    next_page_after_value = None
    if 0 < len(es_response.hits) < es_response.hits.total.value and len(es_response.hits) >= pagination.page_size:
        last = es_response.hits[-1]
        if order_field == doc_type.id_field:
            next_page_after_value = last[doc_type.id_field]
        else:
            # after_value is not necessarily the value stored in the field
            # itself: internally ES can perform the sorting on a different
            # value which is reported under meta.sort.
            after_value = last.meta.sort[0]
            next_page_after_value = '%s:%s' % (after_value, last[doc_type.id_field])
    pagination_response = PaginationResponse(
        total=es_response.hits.total.value,
        next_page_after_value=next_page_after_value,
        **pagination.dict())

    # aggregations
    if len(aggregations) > 0:
        more_response_data['aggregations'] = cast(Dict[str, Any], {
            name: _es_to_api_aggregation(es_response, name, _specific_agg(agg), doc_type=doc_type)
            for name, agg in aggregations.items()})

    more_response_data['es_query'] = es_query.to_dict()
    if isinstance(query, EsQuery):
        # we cannot report EsQuery back, because it won't validate within the MetadataResponse model
        query = None

    result = MetadataResponse(
        owner='all' if owner is None else owner,
        query=query,
        pagination=pagination_response,
        required=required,
        data=[_es_to_entry_dict(hit, required) for hit in es_response.hits],
        **more_response_data)

    return result


def search_iterator(
        owner: str = 'public',
        query: Union[Query, EsQuery] = None,
        order_by: str = 'entry_id',
        required: MetadataRequired = None,
        aggregations: Dict[str, Aggregation] = {},
        user_id: str = None,
        index: Index = entry_index) -> Iterator[Dict[str, Any]]:
    '''
    Works like :func:`search`, but returns an iterator for iterating over the results.
    Consequently, you cannot specify `pagination`, only `order_buy`.
    '''
    page_after_value = None
    while True:
        response = search(
            owner=owner, query=query,
            pagination=MetadataPagination(
                page_size=100, page_after_value=page_after_value, order_by=order_by),
            required=required, aggregations=aggregations, user_id=user_id, index=index)

        page_after_value = response.pagination.next_page_after_value

        for result in response.data:
            yield result

        if page_after_value is None or len(response.data) == 0:
            break


def _index(entries, **kwargs):
    index_entries(entries, **kwargs)


def quantity_values(
        quantity: str, page_size: int = 100, return_buckets: bool = False,
        **kwargs) -> Generator[Any, None, None]:
    '''
    A generator that uses ``search`` and an aggregation to retrieve all
    values of a quantity. Will run multiple requests with page_size until all values
    have been gathered. Kwargs are passed to search, e.g. to change owner or query.
    '''
    page_after_value = None

    while True:
        aggregation = TermsAggregation(quantity=quantity, pagination=AggregationPagination(
            page_size=page_size, page_after_value=page_after_value))

        search_response = search(
            aggregations=dict(value_agg=Aggregation(terms=aggregation)),
            pagination=MetadataPagination(page_size=0),
            **kwargs)

        value_agg = cast(TermsAggregationResponse, search_response.aggregations['value_agg'].terms)  # pylint: disable=no-member
        for bucket in value_agg.data:
            if return_buckets:
                yield bucket
            else:
                yield bucket.value

        if len(value_agg.data) < page_size:
            break

        page_after_value = value_agg.pagination.next_page_after_value
        if page_after_value is None:
            break



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference, MEnum, derived)

from ..common import FastAccess

m_package = Package()


class Atoms(MSection):
    '''
    Describes the atomic structure of the physical system. This includes the atom
    positions, lattice vectors, etc.
    '''

    m_def = Section(validate=False)

    n_atoms = Quantity(
        type=int,
        shape=[],
        description='''
        The total number of species (atoms, particles) in the system.
        ''')

    atomic_numbers = Quantity(
        type=np.dtype(np.int32),
        shape=['n_atoms'],
        description='''
        List of atomic numbers Z of the atoms identified in labels. If a species cannot
        be assigned Z, a negative value can also be used to distinguish it.
        ''')

    equivalent_atoms = Quantity(
        type=np.dtype(np.int32),
        shape=['n_atoms'],
        description='''
        Gives a mapping table of atoms to symmetrically independent atoms in the
        standardized cell. This is used to find symmetrically equivalent atoms.
        ''')

    wyckoff_letters = Quantity(
        type=str,
        shape=['n_atoms'],
        description='''
        Wyckoff letters corresponding to each atom.
        ''')

    concentrations = Quantity(
        type=np.dtype(np.float64),
        shape=['n_atoms'],
        description='''
        Concentrations of the species defined by labels which can be assigned for systems
        with variable compositions.
        ''')

    species = Quantity(
        type=np.dtype(np.int32),
        shape=['n_atoms'],
        description='''
        Species of the atom (normally the atomic number Z, 0 or negative for unidentifed
        species or particles that are not atoms.
        ''')

    labels = Quantity(
        type=str,
        shape=['n_atoms'],
        description='''
        List containing the labels of the atoms. In the usual case, these correspond to
        the chemical symbols of the atoms. One can also append an index if there is a
        need to distinguish between species with the same symbol, e.g., atoms of the
        same species assigned to different atom-centered basis sets or pseudo-potentials,
        or simply atoms in different locations in the structure such as those in the bulk
        and on the surface. In the case where a species is not an atom, and therefore
        cannot be representated by a chemical symbol, the label can simply be the name of
        the particles.
        ''')

    positions = Quantity(
        type=np.dtype(np.float64),
        shape=['n_atoms', 3],
        unit='meter',
        description='''
        Positions of all the species, in cartesian coordinates. This metadata defines a
        configuration and is therefore required. For alloys where concentrations of
        species are given for each site in the unit cell, it stores the position of the
        sites.
        ''')

    velocities = Quantity(
        type=np.dtype(np.float64),
        shape=['n_atoms', 3],
        unit='meter / second',
        description='''
        Velocities of the nuclei, defined as the change in cartesian coordinates of the
        nuclei with respect to time.
        ''')

    lattice_vectors = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        unit='meter',
        description='''
        Lattice vectors in cartesian coordinates of the simulation cell. The
        last (fastest) index runs over the $x,y,z$ Cartesian coordinates, and the first
        index runs over the 3 lattice vectors.
        ''')

    lattice_vectors_reciprocal = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        unit='1/meter',
        description='''
        Reciprocal lattice vectors in cartesian coordinates of the simulation cell. The
        first index runs over the $x,y,z$ Cartesian coordinates, and the second index runs
        over the 3 lattice vectors.
        ''')

    local_rotations = Quantity(
        type=np.dtype(np.float64),
        shape=['n_atoms', 3, 3],
        description='''
        A rotation matrix defining the orientation of each atom. If the rotation matrix
        cannot be specified for an atom, the remaining atoms should set it to
        the zero matrix (not the identity!)
        ''')

    periodic = Quantity(
        type=bool,
        shape=[3],
        description='''
        Denotes if periodic boundary condition is applied to each of the lattice vectors.'
        ''')

    supercell_matrix = Quantity(
        type=np.dtype(np.int32),
        shape=[3, 3],
        description='''
        Specifies the matrix that transforms the unit-cell into the super-cell in which
        the actual calculation is performed.
        ''')

    symmorphic = Quantity(
        type=bool,
        shape=[],
        description='''
        Specifies if the space group is symmorphic. Set to True if all translations are
        zero.
        ''')


class Symmetry(MSection):
    '''
    Section containing information about the symmetry properties of the atomic system.
    '''

    m_def = Section(validate=False)

    bravais_lattice = Quantity(
        type=str,
        shape=[],
        description='''
        Identifier for the Bravais lattice in Pearson notation. The first lowercase letter
        identifies the crystal family and can be one of the following: a (triclinic), b
        (monoclinic), o (orthorhombic), t (tetragonal), h (hexagonal) or c (cubic). The
        second uppercase letter identifies the centring and can be one of the following: P
        (primitive), S (face centred), I (body centred), R (rhombohedral centring) or F
        (all faces centred).
        ''')

    choice = Quantity(
        type=str,
        shape=[],
        description='''
        String that specifies the centering, origin and basis vector settings of the 3D
        space group that defines the symmetry group of the simulated physical system (see
        section system). Values are as defined by spglib.
        ''')

    crystal_system = Quantity(
        type=str,
        shape=[],
        description='''
        Name of the crystal system. Can be one of the following: triclinic, monoclinic,
        orthorhombic, tetragonal, trigonal, hexagonal or cubic.
        ''')

    hall_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        The Hall number for this system.
        ''')

    hall_symbol = Quantity(
        type=str,
        shape=[],
        description='''
        The Hall symbol for this system.
        ''')

    international_short_symbol = Quantity(
        type=str,
        shape=[],
        description='''
        Specifies the International Union of Crystallography (IUC) short symbol of the 3D
        space group of this system
        ''')

    origin_shift = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        description='''
        Vector $\\mathbf{p}$ from the origin of the standardized system to the origin of
        the original system. Together with the matrix $\\mathbf{P}$, found in
        space_group_3D_transformation_matrix, the transformation between the standardized
        coordinates $\\mathbf{x}_s$ and original coordinates $\\mathbf{x}$ is then given
        by $\\mathbf{x}_s = \\mathbf{P} \\mathbf{x} + \\mathbf{p}$.
        ''')

    point_group = Quantity(
        type=str,
        shape=[],
        description='''
        Symbol of the crystallographic point group in the Hermann-Mauguin notation.
        ''')

    space_group_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Specifies the International Union of Crystallography (IUC) number of the 3D space
        group of this system.
        ''')

    symmetry_method = Quantity(
        type=str,
        shape=[],
        description='''
        Identifies the source of the symmetry information contained within this section.
        If equal to 'spg_normalized' the information comes from a normalization step.
        ''')

    transformation_matrix = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        description='''
        Matrix $\\mathbf{P}$ that is used to transform the standardized coordinates to the
        original coordinates. Together with the vector $\\mathbf{p}$, found in
        space_group_3D_origin_shift, the transformation between the standardized
        coordinates $\\mathbf{x}_s$ and original coordinates $\\mathbf{x}$ is then given by
        $\\mathbf{x}_s = \\mathbf{P} \\mathbf{x} + \\mathbf{p}$.
        ''')

    system_original = SubSection(sub_section=Atoms.m_def, repeats=True)

    system_primitive = SubSection(sub_section=Atoms.m_def, repeats=True)

    system_std = SubSection(sub_section=Atoms.m_def, repeats=True)


class Prototype(MSection):
    '''
    Information on the prototype corresponding to the current section.
    '''

    m_def = Section(validate=False)

    aflow_id = Quantity(
        type=str,
        shape=[],
        description='''
        AFLOW id of the prototype (see
        http://aflowlib.org/CrystalDatabase/prototype_index.html) identified on the basis
        of the space_group and normalized_wyckoff.
        ''')

    aflow_url = Quantity(
        type=str,
        shape=[],
        description='''
        Url to the AFLOW definition of the prototype (see
        http://aflowlib.org/CrystalDatabase/prototype_index.html) identified on the basis
        of the space_group and normalized_wyckoff.
        ''')

    assignment_method = Quantity(
        type=str,
        shape=[],
        description='''
        Method used to identify the prototype.
        ''')

    label = Quantity(
        type=str,
        shape=[],
        description='''
        Label of the prototype identified on the basis of the space_group and
        normalized_wyckoff. The label is in the same format as in the read_prototypes
        function: <space_group_number>-<prototype_name>-<Pearson's symbol>).
        ''')


class SpringerMaterial(MSection):
    '''
    Contains results of classification of materials with the same formula according to
    Springer Materials. These include material and compound classsification, formula,
    id, and references from the Springer Materials database.
    '''

    m_def = Section(validate=False)

    id = Quantity(
        type=str,
        shape=[],
        description='''
        Id of the classified material according to Springer Materials.
        ''')

    alphabetical_formula = Quantity(
        type=str,
        shape=[],
        description='''
        The alphabetical formula of the material according to Springer Materials Database
        ''')

    url = Quantity(
        type=str,
        shape=[],
        description='''
        Url to the source page in Springer Materials describing the current entry
        ''')

    compound_class = Quantity(
        type=str,
        shape=['N'],
        description='''
        Name of a class of the current compound, as defined in by Springer Materials. This
        is a property of the chemical formula of the compound
        ''')

    classification = Quantity(
        type=str,
        shape=['N'],
        description='''
        Contains the classification name of the current material according to Springer
        Materials
        ''')


class Constraint(MSection):
    '''
    Section describing a constraint between arbitrary atoms.
    '''

    m_def = Section(validate=False)

    kind = Quantity(
        type=str,
        shape=[],
        description='''
        Short and unique name for this constraint type. Could be fix_a where a can be
        x, y, z denoting a constraint along a cartesian direction, xy, xz, yz denoting a
        constaint along a cartesian plane, xyz denoting a fixed position, distance
        denoting a fixed distance between two atoms, angle denoting a fixed angle between
        three atoms, and dihedral denoting a fixed dihedral angle.
        ''')

    n_constraints = Quantity(
        type=int,
        shape=[],
        description='''
        Number of constraints of this type.
        ''')

    n_atoms = Quantity(
        type=int,
        shape=[],
        description='''
        Number of atoms involved in this constraint.
        ''')

    indices = Quantity(
        type=np.dtype(np.int32),
        shape=['n_constraints', 'n_atoms'],
        description='''
        List of the indexes involved in this constraint.
        ''')

    parameters = Quantity(
        type=typing.Any,
        shape=[],
        description='''
        Explicit constraint parameters for this kind of constraint (depending on the
        constraint type, some might be given implicitly through other means).
        ''')


class System(MSection):
    '''
    Contains parameters describing a system of atomic configuration. These inclue the
    compound name, atomic positions, lattice vectors, contraints on the atoms, etc.
    '''

    m_def = Section(validate=False)

    name = Quantity(
        type=str,
        shape=[],
        description='''
        Specifies the name of the system. This information is provided by the user in some
        codes and is stored here for debugging or visualization purposes.
        ''')

    type = Quantity(
        type=str,
        shape=[],
        description='''
        Type of the system (atom, bulk, surface, etc.) which is determined by the
        normalizer.
        ''')

    configuration_raw_gid = Quantity(
        type=str,
        shape=[],
        description='''
        checksum of the configuration_core, i.e. the geometry of the system. The values
        are not normalized in any way so equivalent configurations might have different
        values
        ''')

    is_representative = Quantity(
        type=bool,
        shape=[],
        description='''
        Most systems in a run are only minor variations of each other. Systems marked
        representative where chosen to be representative for all systems in the run.
        ''')

    n_references = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of references to the current section system.
        ''')

    sub_system_ref = Quantity(
        type=Reference(SectionProxy('System')),
        shape=[],
        description='''
        Links the current section system to a sub system.
        ''',
        categories=[FastAccess])

    systems_ref = Quantity(
        type=Reference(SectionProxy('System')),
        shape=['n_references'],
        description='''
        Links the current section system to other section systems. Such a link is
        necessary for example between the supercell and the reference unit cell in a phonon
        calculation. The relationship should be described by kind and the referred section
        system is given by value. An external url can also be provided in place of value.
        ''',
        categories=[FastAccess])

    atoms = SubSection(sub_section=Atoms.m_def, categories=[FastAccess])

    chemical_composition = Quantity(
        type=str,
        shape=[],
        description='''
        The full chemical composition of the system, based on atom species.
        ''')

    chemical_composition_hill = Quantity(
        type=str,
        shape=[],
        description='''
        The chemical composition in the hill convention of the system, based on atom species.
        ''')

    chemical_composition_reduced = Quantity(
        type=str,
        shape=[],
        description='''
        The chemical composition as reduced formula of the system, based on atom species.
        ''')

    chemical_composition_anonymous = Quantity(
        type=str,
        shape=[],
        description='''
        The chemical composition without explicit element names of the system, based on
        atom species.
        ''')

    constraint = SubSection(sub_section=Constraint.m_def, repeats=True)

    prototype = SubSection(
        sub_section=Prototype.m_def, repeats=True, categories=[FastAccess])

    springer_material = SubSection(
        sub_section=SpringerMaterial.m_def, repeats=True, categories=[FastAccess])

    symmetry = SubSection(
        sub_section=Symmetry.m_def, repeats=True, categories=[FastAccess])


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from nomad.parsing.file_parser import BasicParser
from nomad.units import ureg


class LibAtomsParser:
    def __init__(self):
        re_f = r'\-*\d+\.\d+E*e*\-*\+*\d*'

        self._parser = BasicParser(
            'libAtoms',
            units_mapping=dict(length=ureg.angstrom, energy=ureg.eV),
            # necessary to include version due to auto type conversion
            program_version=r'(svn\_version\=\"\d+\")',
            atom_labels_atom_positions_atom_forces=r'(\<\!\[CDATA\[[A-Z][a-z]* +[\s\S]+?)(?:\<\!\[CDATA\[\d+\]\]\>|\Z)',
            energy_total=rf'slice\_sample energy\=({re_f})'
        )

    def parse(self, mainfile, archive, logger=None):
        self._parser.parse(mainfile, archive, logger=None)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np            # pylint: disable=unused-import
import typing

from pint.util import SharedRegistryObject                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference, MEnum, derived)
from ..common import FastAccess


m_package = Package()


class KMesh(MSection):
    '''
    Section containing the values for
    '''

    m_def = Section(validate=False)

    n_points = Quantity(
        type=int,
        shape=[],
        description='''
        Number of k points in the mesh (i.e. the k points used to evaluate energy_total).
        ''')

    generation_method = Quantity(
        type=str,
        shape=[],
        description='''
        Method used to generate the k points.
        ''')

    density = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='1 / meter ** 3',
        description='''
        Density of k points.
        ''')

    points = Quantity(
        type=np.dtype(np.float64),
        shape=['n_points', 3],
        description='''
        List of all the k points in the $k$-point mesh. These are the k point used to
        evaluate energy_total, and are in fractional coordinates (in the basis of the
        reciprocal-lattice vectors).
        ''')

    weights = Quantity(
        type=np.dtype(np.float64),
        shape=['n_points'],
        description='''
        Weights of all the k points in the $k$-point mesh. These are the weights for
        k_mesh_points (i.e. the k point used to evaluate energy_total).
        ''')


class Scf(MSection):
    '''
    Section containing the parameters related to self consistency.
    '''

    m_def = Section(validate=False)

    n_max_iteration = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Specifies the maximum number of allowed self-consistent field (SCF) iterations in
        a calculation.
        ''')

    threshold_energy_change = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Specifies the threshold for the total energy change between two subsequent
        self-consistent field (SCF) iterations. The SCF is considered converged when the
        total-energy change between two SCF cycles is below the threshold (possibly in
        combination with other criteria).
        ''')

    threshold_density_change = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Specifies the threshold for the average  charge density change between two
        subsequent self-consistent field (SCF) iterations. The SCF is considered converged
        when the density change between two SCF cycles is below the threshold (possibly in
        combination with other criteria).
        ''')

    minimization_algorithm = Quantity(
        type=str,
        shape=[],
        description='''
        Specifies the algorithm used for self consistency minimization.
        ''')


class AtomParameters(MSection):
    '''
    Contains method-related information about a kind of atom identified by label. This
    allows the assignment of an atom-centered basis set or pseudopotential for different
    atoms belonging to the same kind.
    '''

    m_def = Section(validate=False)

    atom_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Atomic number (number of protons) of this atom kind, use 0 if not an atom.
        ''')

    n_valence_electrons = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number of valence electrons.
        ''')

    label = Quantity(
        type=str,
        shape=[],
        description='''
        String used to identify the atoms of this kind. This should correspond to the
        atom labels of the configuration. It is possible for one atom kind to have
        multiple labels (in order to allow two atoms of the same kind to have two
        differently defined sets of atom-centered basis functions or two different pseudo-
        potentials). Atom kind is typically the symbol of the atomic species but it can be
        also a ghost or pseudo-atom.
        ''')

    mass = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='kg',
        description='''
        Mass of the atom.
        ''')

    pseudopotential_name = Quantity(
        type=str,
        shape=[],
        description='''
        Name identifying the pseudopotential used.
        ''')

    n_orbitals = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of valence orbitals of the atom.
        ''')

    orbitals = Quantity(
        type=str,
        shape=['n_orbitals'],
        description='''
        Label of the valence orbitals of the atoms.
        ''')

    onsite_energies = Quantity(
        type=np.dtype(np.float64),
        shape=['n_orbitals'],
        unit='joule',
        description='''
        Values of the atomic onsite energy corresponding to each orbital.
        ''')

    charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='coulomb',
        description='''
        Total charge of the atom.
        ''')

    charges = Quantity(
        type=np.dtype(np.float64),
        shape=['n_orbitals'],
        unit='coulomb',
        description='''
        Values of the charge corresponding to each orbital.
        ''')


class MoleculeParameters(MSection):
    '''
    Contains method-related information about a kind of atom identified by label. This
    allows the assignment of an atom-centered basis set or pseudopotential for different
    atoms belonging to the same kind.
    '''

    m_def = Section(validate=False)

    label = Quantity(
        type=str,
        shape=[],
        description='''
        String to identify the molecule.
        ''')

    n_atoms = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of atoms in the molecule.
        ''')

    atoms = SubSection(sub_section=AtomParameters.m_def, repeats=True)


class GaussianBasisGroup(MSection):
    '''
    Section that describes a group of Gaussian contractions. Groups allow one to calculate
    the primitive Gaussian integrals once for several different linear combinations of
    them. This defines basis functions with radial part $f_i(r) = r^{l_i} \\sum_{j} c_{i j}
    A(l_i, \\alpha_j) exp(-\\alpha_j r^2)$ where $A(l_i, \\alpha_j)$ is a the normalization
    coefficient for primitive Gaussian basis functions. Here, $\\alpha_j$ is defined in
    gaussian_basis_group_exponents, $l_i$ is given in gaussian_basis_group_ls, and $c_{i
    j}$ is given in gaussian_basis_group_contractions, whereas the radial part is given by
    the spherical harmonics $Y_{l m}$.

    This section is defined only if the original basis function uses Gaussian basis
    functions, and the sequence of radial functions $f_i$ across all
    section_gaussian_basis_group in section_basis_set_atom_centered should match the one
    of basis_set_atom_centered_radial_functions.
    '''

    m_def = Section(validate=False)

    n_contractions = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of different contractions, i.e. resulting basis functions in a
        gaussian_basis_group section.
        ''')

    n_exponents = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of different Gaussian exponents in a section_gaussian_basis_group
        section.
        ''')

    contractions = Quantity(
        type=np.dtype(np.float64),
        shape=['n_contractions', 'n_exponents'],
        description='''
        contraction coefficients $c_{i j}$ defining the contracted basis functions with
        respect to *normalized* primitive Gaussian functions. They define the Gaussian
        basis functions as described in section_gaussian_basis_group.
        ''')

    exponents = Quantity(
        type=np.dtype(np.float64),
        shape=['n_exponents'],
        unit='1 / meter ** 2',
        description='''
        Exponents $\\alpha_j$ of the Gaussian functions defining this basis set
        $exp(-\\alpha_j r^2)$. One should be careful about the units of the coefficients.
        ''')

    ls = Quantity(
        type=np.dtype(np.float64),
        shape=['n_contractions'],
        description='''
        Azimuthal quantum number ($l$) values (of the angular part given by the spherical
        harmonic $Y_{l m}$ of the various contracted basis functions).
        ''')


class BasisSetAtomCentered(MSection):
    '''
    This section describes the atom-centered basis set. The main contained information is
    a short, non unique but human-interpretable, name for identifying the basis set
    (short_name), a longer unique name, the atomic number of the atomic species the
    basis set is meant for.
    '''

    m_def = Section(validate=False)

    name = Quantity(
        type=str,
        shape=[],
        description='''
        Code-specific, but explicative, base name for the basis set.
        ''')

    atom_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Atomic number (i.e., number of protons) of the atom for which this basis set is
        constructed (0 means unspecified or a pseudo atom).
        ''')

    n_basis_functions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Gives the number of different basis functions in a basis_set_atom_centered
        section. This equals the number of actual coefficients that are specified when
        using this basis set.
        ''')

    gaussian_basis_group = SubSection(sub_section=GaussianBasisGroup.m_def, repeats=True)


class BasisSetCellDependent(MSection):
    '''
    Section describing a cell-dependent (atom-independent) basis set, e.g. plane waves.
    The contained information is the type of basis set (in basis_set_cell_dependent_kind),
    its parameters (e.g., for plane waves in basis_set_planewave_cutoff), and a name that
    identifies the actually used basis set (a string combining the type and the
    parameter(s), stored in name).
    '''

    m_def = Section(validate=False)

    kind = Quantity(
        type=str,
        shape=[],
        description='''
        A string defining the type of the cell-dependent basis set (i.e., non atom
        centered such as plane-waves). Can be one of plane_waves, realspace_grids or
        wavelets.
        ''')

    name = Quantity(
        type=str,
        shape=[],
        description='''
        A label identifying the cell-dependent basis set (i.e., non atom centered such as
        plane-waves). The following convetion should be followed:
        plane_waves ("PW_" + cutoff in Ry) realspace_grids ("GR_" + grid spacing in fm)
        wavelets (WV_" + smallest wavelet spacing in fm).
        ''')

    planewave_cutoff = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Spherical cutoff  in reciprocal space for a plane-wave basis set. It is the energy
        of the highest plan-ewave ($\\frac{\\hbar^2|k+G|^2}{2m_e}$) included in the basis
        set. Note that normally this basis set is used for the wavefunctions, and the
        density would have 4 times the cutoff, but this actually depends on the use of the
        basis set by the method.
        ''')

    grid_spacing = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        Grid spacing used for the realspace representation of the wave functions.
        ''')


class BasisSet(MSection):
    '''
    This section contains all basis sets used to represent the wavefunction or electron
    density.
    '''

    m_def = Section(validate=False)

    kind = Quantity(
        type=str,
        shape=[],
        description='''
        String describing the use of the basis set, i.e, if it used for expanding a
        wavefunction or an electron density.
        ''')

    type = Quantity(
        type=str,
        shape=[],
        description='''
        The type of basis set used by the program. Valid values are: [`Numeric AOs`,
        `Gaussians`, `(L)APW+lo`, `plane waves`, `psinc functions`, `real-space grid`].
        ''')

    name = Quantity(
        type=str,
        shape=[],
        description='''
        Identifies the basis set.
        ''')

    cell_dependent = SubSection(sub_section=BasisSetCellDependent.m_def, repeats=True)

    atom_centered = SubSection(sub_section=BasisSetAtomCentered.m_def, repeats=True)


class Interaction(MSection):
    '''
    Section containing the parameters of a contribution to a force field model.
    '''

    m_def = Section(validate=False)

    type = Quantity(
        type=str,
        shape=[],
        description='''
        Denotes the classification of the potential. Could be one of pair, LJ, Morse, EAM,
        MEAM.
        ''')

    name = Quantity(
        type=str,
        shape=[],
        description='''
        Specifies the name of the potential. Can contain information on the species,
        cut-offs, potential versions.
        ''')

    atom_labels = Quantity(
        type=str,
        shape=[2],
        description='''
        Labels of the atoms described by the interaction.
        ''')

    atom_indices = Quantity(
        type=np.dtype(np.int32),
        shape=[2],
        description='''
        Indices of the atoms in the system described by the interaction.
        ''')

    functional_form = Quantity(
        type=str,
        shape=[],
        description='''
        Specifies the functional form of the interaction potential.
        ''')

    n_parameters = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Specifies the number of parameters in the interaction potential.
        ''')

    parameters = Quantity(
        type=typing.Any,
        shape=[],
        description='''
        Dictionary of label and parameters of the interaction potential.
        ''')


class Model(MSection):
    '''
    Section containing the parameters of a force field model. If specified, the parameters
    corresponding to the individual contributions to the model are given in contributions.
    Otherwise, the parameters can also be found in a reference to the published model.
    '''

    m_def = Section(validate=False)

    name = Quantity(
        type=str,
        shape=[],
        description='''
        Identifies the name of the model.
        ''')

    reference = Quantity(
        type=str,
        shape=[],
        description='''
        Reference to the model e.g. DOI, URL.
        ''')

    contributions = SubSection(sub_section=Interaction.m_def, repeats=True)


class Functional(MSection):
    '''
    Section containing the parameters of an exchange or correlation functional.
    '''

    m_def = Section(validate=False)

    name = Quantity(
        type=str,
        shape=[],
        description='''
        Provides the name of one of the exchange and/or correlation (XC) functional
        following the libbx convention.
        ''')

    parameters = Quantity(
        type=typing.Any,
        shape=[],
        description='''
        Contains an associative list of non-default values of the parameters for the
        functional.

        For example, if a calculations using a hybrid XC functional (e.g., HSE06)
        specifies a user-given value of the mixing parameter between exact and GGA
        exchange, then this non-default value is stored in this metadata.

        The labels and units of these values may be defined in name.

        If this metadata is not given, the default parameter values for the functional
        are assumed.
        ''')

    weight = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Provides the value of the weight for the functional.

        This weight is used in the linear combination of the different functionals. If not
        specified then the default is set to 1.
        ''')


class XCFunctional(Model):
    '''
    Section describing the exchange-correlation functional used in the DFT calculation.
    The name of the exchange-correlation functional is given by name and the reference to
    the published functional is provided by reference. Other contributions to the
    functional not covered by exchange, correlation or hybrid types may be specified in
    contributions.
    '''

    m_def = Section(validate=False)

    exchange = SubSection(sub_section=Functional.m_def, repeats=True)

    correlation = SubSection(sub_section=Functional.m_def, repeats=True)

    hybrid = SubSection(sub_section=Functional.m_def, repeats=True)

    contributions = SubSection(sub_section=Functional.m_def, repeats=True)


class DFTPlusU(MSection):
    '''
    Section for DFT+U-settings of a single orbital
    '''

    m_def = Section(validate=False)

    atom = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Index of the atom corresponding to the parameter setting
        ''')

    orbital = Quantity(
        type=str,
        shape=[],
        description='''
        Orbital label corresponding to the parameter setting following the notation:
        '3d', '4f', ...
        ''')

    U_effective = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Value of the effective U parameter (U-J).
        ''')

    U = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Value of the on-site Coulomb interaction U
        ''')

    J = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Value of the exchange interaction J
        ''')

    functional = Quantity(
        type=str,
        shape=[],
        description='''
        Type of DFT+U functional (such as DFT/DFT+U double-counting compensation).
        ''')

    projection_type = Quantity(
        type=str,
        shape=[],
        description='''
        DFT+U: Type of orbitals used for projection in order to calculate occupation
        numbers.
        ''')


class DFT(MSection):
    '''
    Section containing the various parameters that define a DFT calculation. These include
    settings for the exchange correlation functionals, LDA+U, etc.
    '''

    m_def = Section(validate=False)

    self_interaction_correction_method = Quantity(
        type=str,
        shape=[],
        description='''
        Contains the name for the self-interaction correction (SIC) treatment used to
        calculate the final energy and related quantities. If skipped or empty, no special
        correction is applied.

        The following SIC methods are available:

        | SIC method                | Description                       |

        | ------------------------- | --------------------------------  |

        | `""`                      | No correction                     |

        | `"SIC_AD"`                | The average density correction    |

        | `"SIC_SOSEX"`             | Second order screened exchange    |

        | `"SIC_EXPLICIT_ORBITALS"` | (scaled) Perdew-Zunger correction explicitly on a
        set of orbitals |

        | `"SIC_MAURI_SPZ"`         | (scaled) Perdew-Zunger expression on the spin
        density / doublet unpaired orbital |

        | `"SIC_MAURI_US"`          | A (scaled) correction proposed by Mauri and co-
        workers on the spin density / doublet unpaired orbital |
        ''')

    xc_functional = SubSection(sub_section=XCFunctional.m_def)

    dft_plus_u = SubSection(sub_section=DFTPlusU.m_def, repeats=True)


class GW(MSection):
    '''
    Section containing the various parameters that define a GW calculation.
    '''

    m_def = Section(validate=False)

    bare_coulomb_cutofftype = Quantity(
        type=str,
        shape=[],
        description='''
        Cutoff type for the calculation of the bare Coulomb potential: none, 0d, 1d, 2d.
        See Rozzi et al., PRB 73, 205119 (2006)
        ''')

    bare_coulomb_gmax = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='1 / meter',
        description='''
        Maximum G for the pw basis for the Coulomb potential.
        ''')

    basis_set = Quantity(
        type=str,
        shape=[],
        description='''
        Auxillary basis set used for non-local operators: mixed - mixed basis set, Kotani
        and Schilfgaarde, Solid State Comm. 121, 461 (2002).
        ''')

    core_treatment = Quantity(
        type=str,
        shape=[],
        description='''
        It specifies whether the core states are treated in the GW calculation: all - All
        electron calculation; val - Valence electron only calculation; vab - Core
        electrons are excluded from the mixed product basis; xal - All electron treatment
        of the exchange self-energy only
        ''')

    frequency_grid_type = Quantity(
        type=str,
        shape=[],
        description='''
        Frequency integration grid type for the correlational self energy: 'eqdis' -
        equidistant frequencies from 0 to freqmax; 'gaulag' - Gauss-Laguerre quadrature
        from 0 to infinity; 'gauleg' - Gauss-Legendre quadrature from 0 to freqmax;
        'gaule2' (default) - double Gauss-Legendre quadrature from 0 to freqmax and from
        freqmax to infinity.
        ''')

    max_frequency = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Maximum frequency for the calculation of the self energy.
        ''')

    mixed_basis_gmax = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='1 / meter',
        description='''
        Cut-off parameter for the truncation of the expansion of the plane waves in the
        interstitial region.
        ''')

    mixed_basis_lmax = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Maximum l value used for the radial functions within the muffin-tin.
        ''')

    mixed_basis_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Eigenvalue threshold below which the egenvectors are discarded in the construction
        of the radial basis set.
        ''')

    ngridq = Quantity(
        type=np.dtype(np.int32),
        shape=[3],
        description='''
        k/q-point grid size used in the GW calculation.
        ''')

    n_frequencies = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of frequency points used in the calculation of the self energy.
        ''')

    frequency_number = Quantity(
        type=np.dtype(np.int32),
        shape=['n_frequencies'],
        description='''
        Number referring to the frequency used in the calculation of the self energy.
        ''')

    frequency_values = Quantity(
        type=np.dtype(np.float64),
        shape=['n_frequencies'],
        unit='joule',
        description='''
        Values of the frequency used in the calculation of the self energy.
        ''')

    frequency_weights = Quantity(
        type=np.dtype(np.float64),
        shape=['n_frequencies'],
        description='''
        Weights of the frequency used in the calculation of the self energy.
        ''')

    polarizability_number_of_empty_states = Quantity(
        type=int,
        shape=[],
        description='''
        Number of empty states used to compute the polarizability P
        ''')

    qp_equation_treatment = Quantity(
        type=str,
        shape=[],
        description='''
        Methods to solve the quasi-particle equation: 'linearization', 'self-consistent'
        ''')

    screened_coulomb_volume_average = Quantity(
        type=str,
        shape=[],
        description='''
        Type of volume averaging for the dynamically screened Coulomb potential: isotropic
        - Simple averaging along a specified direction using only diagonal components of
        the dielectric tensor; anisotropic - Anisotropic screening by C. Freysoldt et al.,
        CPC 176, 1-13 (2007)
        ''')

    screened_coulomb = Quantity(
        type=str,
        shape=[],
        description='''
        Model used to calculate the dinamically-screened Coulomb potential: 'rpa' - Full-
        frequency random-phase approximation; 'ppm' - Godby-Needs plasmon-pole model Godby
        and Needs, Phys. Rev. Lett. 62, 1169 (1989); 'ppm_hl' - Hybertsen and Louie, Phys.
        Rev. B 34, 5390 (1986); 'ppm_lh' - von der Linden and P. Horsh, Phys. Rev. B 37,
        8351 (1988); 'ppm_fe' - Farid and Engel, Phys. Rev. B 47,15931 (1993); 'cdm' -
        Contour deformation method, Phys. Rev. B 67, 155208 (2003).)
        ''')

    self_energy_c_analytical_continuation = Quantity(
        type=str,
        shape=[],
        description='''
        Models for the correlation self-energy analytical continuation: 'pade' -  Pade's
        approximant (by H. J. Vidberg and J. W. Serence, J. Low Temp. Phys. 29, 179
        (1977)); 'mpf' -  Multi-Pole Fitting (by H. N Rojas, R. W. Godby and R. J. Needs,
        Phys. Rev. Lett. 74, 1827 (1995)); 'cd' - contour deformation; 'ra' - real axis
        ''')

    self_energy_c_number_of_empty_states = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of empty states to be used to calculate the correlation self energy.
        ''')

    self_energy_c_number_of_poles = Quantity(
        type=int,
        shape=[],
        description='''
        Number of poles used in the analytical continuation.
        ''')

    self_energy_singularity_treatment = Quantity(
        type=str,
        shape=[],
        description='''
        Treatment of the integrable singular terms in the calculation of the self energy.
        Values: 'mpb' - Auxiliary function method by S. Massidda, M. Posternak, and A.
        Baldereschi, PRB 48, 5058 (1993); 'crg' - Auxiliary function method by P. Carrier,
        S. Rohra, and A. Goerling, PRB 75, 205126 (2007).
        ''')

    starting_point = Quantity(
        type=str,
        shape=[],
        description='''
        Exchange-correlation functional of the ground-state calculation.
        ''')

    type = Quantity(
        type=MEnum(["G0W0", "scGW", "scGW0", "scG0W", "ev-scGW", "qp-scGW"]),
        shape=[],
        description='''
        GW methodology: G0W0; ev-scGW: (eigenvalues self-consistent GW) – Phys.Rev.B 34,
        5390 (1986); qp-scGW: (quasi-particle self-consistent GW) – Phys. Rev. Lett. 96,
        226402 (2006)  scGW0: (self-consistent G with fixed W0) – Phys.Rev.B 54, 8411
        (1996); scG0W: (self-consistent W with fixed G0); scGW: (self-consistent GW) –
        Phys. Rev. B 88, 075105 (2013)
        ''')


class TBModel(Model):
    '''
    Section containing the parameters pertaining to a tight-binding calculation.
    '''

    m_def = Section(validate=False)

    hamiltonian = SubSection(sub_section=Interaction.m_def, repeats=True)

    overlap = SubSection(sub_section=Interaction.m_def, repeats=True)

    repulsion = SubSection(sub_section=Interaction.m_def, repeats=True)

    magnetic = SubSection(sub_section=Interaction.m_def, repeats=True)

    coulomb = SubSection(sub_section=Interaction.m_def, repeats=True)


class TB(MSection):
    '''
    Section containing the parameters pertaining to a tight-binding calculation.
    '''

    m_def = Section(validate=False)

    kind = Quantity(
        type=str,
        shape=[],
        description='''
        The kind of tight-binding model used. Can be orthogonal or non-orthogonal.
        ''')

    model = SubSection(sub_section=TBModel.m_def, repeats=True)


class ForceField(MSection):
    '''
    Section containing the parameters pertaining to a force field calculation.
    '''

    m_def = Section(validate=False)

    model = SubSection(sub_section=Model.m_def, repeats=True)


class Smearing(MSection):
    '''
    Section containing the parameters related to the smearing of the electronic density of
    states at the Fermi level.
    '''

    m_def = Section(validate=False)

    kind = Quantity(
        type=str,
        shape=[],
        description='''
        Specifies the kind of smearing on the electron occupation used to calculate the
        free energy (see energy_free)

        Valid values are:

        | Smearing kind             | Description                       |

        | ------------------------- | --------------------------------- |

        | `"empty"`                 | No smearing is applied            |

        | `"gaussian"`              | Gaussian smearing                 |

        | `"fermi"`                 | Fermi smearing                    |

        | `"marzari-vanderbilt"`    | Marzari-Vanderbilt smearing       |

        | `"methfessel-paxton"`     | Methfessel-Paxton smearing        |

        | `"tetrahedra"`            | Interpolation of state energies and occupations
        (ignores smearing_width) |
        ''')

    width = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Specifies the width of the smearing in energy for the electron occupation used to
        calculate the free energy (see energy_free).

        *NOTE:* Not all methods specified in smearing_kind uses this value.
        ''')


class Electronic(MSection):
    '''
    Section containing the parameters related to the electronic structure.
    '''

    m_def = Section(validate=False)

    spin_target = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Stores the target (user-imposed) value of the spin multiplicity $M=2S+1$, where
        $S$ is the total spin. It is an integer number. This value is not necessarily the
        value obtained at the end of the calculation. See spin_S2 for the converged value
        of the spin moment.
        ''')

    charge = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        unit='coulomb',
        description='''
        Stores the total charge of the system.
        ''')

    n_bands = Quantity(
        type=int,
        shape=[],
        description='''
        Specifies the number of bands used in the calculation.
        ''')

    n_spin_channels = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of spin channels.
        ''')

    n_electrons = Quantity(
        type=np.dtype(np.float64),
        shape=['n_spin_channels'],
        description='''
        Number of electrons in system
        ''')

    method = Quantity(
        type=str,
        shape=[],
        description='''
        Non-unique string identifying the used electronic structure method. It is not
        unique in the sense that two calculations with the same
        electronic structure method string may have not been performed with exactly the
        same method.
        ''')

    relativity_method = Quantity(
        type=MEnum(['scalar_relativistic', 'pseudo_scalar_relativistic', 'scalar_relativistic_atomic_ZORA']),
        shape=[],
        description='''
        Describes the relativistic treatment used for the calculation of the final energy
        and related quantities. If skipped or empty, no relativistic treatment is applied.
        ''')

    van_der_waals_method = Quantity(
        type=str,
        shape=[],
        description='''
        Describes the Van der Waals method. If skipped or an empty string is used, it
        means no Van der Waals correction is applied.

        Allowed values are:

        | Van der Waals method  | Description                               |

        | --------------------- | ----------------------------------------- |

        | `""`                  | No Van der Waals correction               |

        | `"TS"`                | A. Tkatchenko and M. Scheffler, [Phys. Rev. Lett.
        **102**, 073005 (2009)](http://dx.doi.org/10.1103/PhysRevLett.102.073005) |

        | `"OBS"`               | F. Ortmann, F. Bechstedt, and W. G. Schmidt, [Phys. Rev.
        B **73**, 205101 (2006)](http://dx.doi.org/10.1103/PhysRevB.73.205101) |

        | `"G06"`               | S. Grimme, [J. Comput. Chem. **27**, 1787
        (2006)](http://dx.doi.org/10.1002/jcc.20495) |

        | `"JCHS"`              | P. Jurečka, J. Černý, P. Hobza, and D. R. Salahub,
        [Journal of Computational Chemistry **28**, 555
        (2007)](http://dx.doi.org/10.1002/jcc.20570) |

        | `"MDB"`               | Many-body dispersion. A. Tkatchenko, R. A. Di Stasio Jr,
        R. Car, and M. Scheffler, [Physical Review Letters **108**, 236402
        (2012)](http://dx.doi.org/10.1103/PhysRevLett.108.236402) and A. Ambrosetti, A. M.
        Reilly, R. A. Di Stasio Jr, and A. Tkatchenko, [The Journal of Chemical Physics
        **140**, 18A508 (2014)](http://dx.doi.org/10.1063/1.4865104) |

        | `"XC"`                | The method to calculate the Van der Waals energy uses a
        non-local functional which is described in section_XC_functionals. |
        ''')

    smearing = SubSection(sub_section=Smearing.m_def)


class Method(MSection):
    '''
    Section containing the various parameters that define the theory and the
    approximations (convergence, thresholds, etc.) behind the calculation.
    '''

    m_def = Section(validate=False)

    stress_tensor_method = Quantity(
        type=str,
        shape=[],
        description='''
        Specifies the method used to calculate stress_tensor for, e.g., molecular dynamics
        and geometry optimization.

        The allowed values are:

        * numeric

        * analytic
        ''')

    starting_method_ref = Quantity(
        type=Reference(SectionProxy('Method')),
        shape=[],
        description='''
        Links the current section method to a section method containing the starting
        parameters.
        ''',
        categories=[FastAccess])

    core_method_ref = Quantity(
        type=Reference(SectionProxy('Method')),
        shape=[],
        description='''
        Links the current section method to a section method containing the core settings.
        ''',
        categories=[FastAccess])

    n_references = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of references to the current method.
        ''')

    methods_ref = Quantity(
        type=Reference(SectionProxy('Method')),
        shape=['n_references'],
        description='''
        Links the section method to other method sections. For instance, one calculation
        is a perturbation performed using a self-consistent field (SCF) calculation as
        starting point, or a simulated system is partitioned in regions with different but
        connected Hamiltonians (e.g., QM/MM, or a region treated via Kohn-Sham DFT
        embedded into a region treated via orbital-free DFT).
        ''',
        categories=[FastAccess])

    dft = SubSection(sub_section=DFT.m_def)

    gw = SubSection(sub_section=GW.m_def)

    tb = SubSection(sub_section=TB.m_def)

    force_field = SubSection(sub_section=ForceField.m_def)

    k_mesh = SubSection(sub_section=KMesh.m_def)

    electronic = SubSection(sub_section=Electronic.m_def)

    scf = SubSection(sub_section=Scf.m_def)

    atom_parameters = SubSection(sub_section=AtomParameters.m_def, repeats=True, label_quantity='label')

    molecule_parameters = SubSection(sub_section=MoleculeParameters.m_def, repeats=True, label_quantity='label')

    basis_set = SubSection(sub_section=BasisSet.m_def, repeats=True, label_quantity='type')


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import

from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference, JSON
)
from nomad.datamodel.metainfo import simulation


m_package = Package()


class AtomParameters(simulation.method.AtomParameters):

    m_def = Section(validate=False, extends_base_section=True)

    atom = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''')

    x_dftbp_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')


class System(simulation.system.System):

    m_def = Section(validate=False, extends_base_section=True)

    x_dftbp_atom_positions_X = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')

    x_dftbp_atom_positions_Y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')

    x_dftbp_atom_positions_Z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')


class BandEnergies(simulation.calculation.BandEnergies):

    m_def = Section(validate=False, extends_base_section=True)

    x_dftbp_eigenvalues_values = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')

    x_dftbp_eigenvalues_occupation = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')


class Calculation(simulation.calculation.Calculation):

    m_def = Section(validate=False, extends_base_section=True)

    x_dftbp_atom_forces_X = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')

    x_dftbp_atom_forces_Y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')

    x_dftbp_atom_forces_Z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')

    x_dftbp_force_max = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')

    x_dftbp_force_max_mov = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''')


class Energy(simulation.calculation.Energy):

    m_def = Section(validate=False, extends_base_section=True)

    x_dftbp_total_mermin = SubSection(simulation.calculation.EnergyEntry.m_def)

    x_dftbp_band = SubSection(simulation.calculation.EnergyEntry.m_def)

    x_dftbp_ts = SubSection(simulation.calculation.EnergyEntry.m_def)

    x_dftbp_band_free = SubSection(simulation.calculation.EnergyEntry.m_def)

    x_dftbp_band_t0 = SubSection(simulation.calculation.EnergyEntry.m_def)

    x_dftbp_scc = SubSection(simulation.calculation.EnergyEntry.m_def)

    x_dftbp_dispersion = SubSection(simulation.calculation.EnergyEntry.m_def)

    x_dftbp_force_related = SubSection(simulation.calculation.EnergyEntry.m_def)


class Method(simulation.method.Method):

    m_def = Section(validate=False, extends_base_section=True)

    x_dftbp_input_parameters = Quantity(
        type=JSON,
        shape=[],
        description='''
        -
        ''')


class TB(simulation.method.TB):

    m_def = Section(validate=False, extends_base_section=True)

    x_dftbp_n_sk_files = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_dftbp_sk_files = Quantity(
        type=str,
        shape=['x_dftbp_n_sk_files'],
        description='''
        ''')


class Run(simulation.run.Run):

    m_def = Section(validate=False, extends_base_section=True)

    x_dftbp_input_file = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_dftbp_processed_input_file = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_dftbp_parser_version = Quantity(
        type=str,
        shape=[],
        description='''
        ''')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import logging
import numpy as np

from nomad.units import ureg
from nomad.parsing.file_parser import TextParser, Quantity
from nomad.datamodel.metainfo.simulation.run import (
    Run, Program
)
from nomad.datamodel.metainfo.simulation.system import (
    System, Atoms
)
from nomad.datamodel.metainfo.simulation.method import (
    Method, BasisSet, DFT, XCFunctional, Functional, Electronic, Smearing
)
from nomad.datamodel.metainfo.simulation.calculation import (
    Calculation, ScfIteration, Energy, EnergyEntry, Forces, ForcesEntry, BandEnergies
)
from .metainfo.fleur import x_fleur_header


re_n = r'[\n\r]'
re_f = r'[-+]?\d+\.\d*(?:[DdEe][-+]\d+)?'


class XMLParser(TextParser):
    def init_quantities(self):
        self._quantities = [
            Quantity(
                'header',
                r'(\<programVersion[\s\S]+?\<\/programVersion\>',
                sub_parser=TextParser(quantities=[
                    Quantity('program_version', r'programVersion version=\"fleur (\S+)\"', dtype=str),
                    Quantity('x_fleur_precision', r'precision type=\"\S+\"', dtype=str),
                    Quantity(
                        'x_fleur_structure_class',
                        r'\<targetStructureClass\>(.+?)\<\/targetStructureClass\>',)
                ])
            ),
        ]


class OutParser(TextParser):
    def init_quantities(self):
        def to_atoms(val_in):
            labels, positions = [], []
            for val in val_in.strip().splitlines():
                val = val.split()
                labels.append(val[1])
                positions.append(val[6:9])
            return labels, np.array(positions, dtype=np.float64) * ureg.bohr

        def to_key_val(val_in):
            val = val_in.strip().replace('D', 'E').split('=')
            key, val = val[0].strip(), val[1].split()
            return [key, val[0] if len(val) == 1 else val]

        key_val_quantity = Quantity(
            'key_val',
            r'([\w \-\/]+\s*=\s*[\d\.\-\+\w ]+)',
            str_operation=to_key_val, repeats=True)

        self._quantities = [
            Quantity(
                'header',
                r'(This output is generated[\s\S]+?Additional flags.+)',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'program_version',
                        r'This output is generated by (fleur\S+)',
                        dtype=str
                    ),
                    Quantity('x_fleur_precision', r'(\S+) precision', dtype=str),
                    Quantity('x_fleur_with_inversion', r'(\S+) INVERSION', dtype=lambda x: x == 'with'),
                    Quantity('x_fleur_with_soc', r'(\S+) SOC', dtype=lambda x: x == 'with'),
                    Quantity(
                        'x_fleur_additional_flags',
                        r'Additional flags are: +(.+)',
                        str_operation=lambda x: x.strip(), dtype=str)
                ])
            ),
            Quantity(
                'input',
                r'(Your parameters\:[\s\S]+?fleur)',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'input_parameters',
                        r'dump of inp-file +\-+([\s\S]+?)Local',
                        sub_parser=TextParser(quantities=[key_val_quantity])
                    ),
                    Quantity(
                        'parameters',
                        r'(parameter *\([\s\S]+?)fleur',
                        sub_parser=TextParser(quantities=[key_val_quantity])
                    )
                ])
            ),
            Quantity(
                'system',
                r'([\w ]+\s+lattice=[\s\S]+?)input of',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'parameters',
                        rf'(lattice[\s\S]+?){re_n} *{re_n}',
                        sub_parser=TextParser(quantities=[key_val_quantity])
                    ),
                    Quantity(
                        'x_fleur_unit_cell_volume',
                        rf'the volume of the unit cell omega\-tilda= +({re_f})',
                        dtype=np.float64
                    ),
                    Quantity(
                        'x_fleur_unit_cell_volume_omega',
                        rf'the volume of the unit cell omega= +({re_f})',
                        dtype=np.float64
                    ),
                    Quantity('exchange_correlation', r'exchange\-correlation: (.+)', flatten=False, dtype=str),
                    Quantity('x_fleur_k_max', rf'k_max *= *({re_f})', dtype=np.float64),
                    Quantity('x_fleur_G_max', rf'G_max *= *({re_f})', dtype=np.float64),
                    Quantity('x_fleur_vol_interstitial', rf'volume of interstitial region= +({re_f})', dtype=np.float64),
                    Quantity('x_fleur_nr_of_atom_types', rf'number of atom types= +(\d+)', dtype=np.int32),
                    Quantity(
                        'cell',
                        r'bravais matrices of real and reciprocal lattices\s+'
                        # rf'((?:{re_f} +{re_f} +{re_f} +{re_f} +{re_f} +{re_f}\s+))',
                        rf'({re_f} +{re_f} +{re_f}) +{re_f} +{re_f} +{re_f}\s+'
                        rf'({re_f} +{re_f} +{re_f}) +{re_f} +{re_f} +{re_f}\s+'
                        rf'({re_f} +{re_f} +{re_f}) +{re_f} +{re_f} +{re_f}\s+',
                        dtype=np.dtype(np.float64), shape=(3, 3), unit=ureg.bohr
                    ),
                    Quantity(
                        'atoms',
                        r'no\. +type +int\.\-coord\. +cart\.coord\. +rmt +jri +dx +lmax\s+'
                        rf'((?:\d+ +[A-Z][a-z] +\d+.+\s+)+)',
                        str_operation=to_atoms
                    ),
                ])
            ),
            Quantity(
                'electronic',
                r'(parameters for eigenvalues[\s\S]+?)Local',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'eigenvalues_parameters',
                        rf'parameters for eigenvalues:([\s\S]+?){re_n} *{re_n}',
                        sub_parser=TextParser(quantities=[key_val_quantity])
                    ),
                    Quantity(
                        'smearing',
                        r'gauss\-integration is used += +(\S)',
                        str_operation=lambda x: 'gaussian' if x == 'T' else 'fermi'
                    ),
                    Quantity('width', rf'(?:width|broadening) += +({re_f})', dtype=np.float64),
                    Quantity(
                        'x_fleur_nr_of_valence_electrons',
                        rf'number of valence electrons= +({re_f})',
                        dtype=np.float64
                    )

                ])
            ),
            Quantity(
                'electic_field_parameters',
                rf'parameters for external electric field:([\s\S]+?){re_n} *{re_n}',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'x_fleur_tot_elec_charge',
                        rf'total electronic charge += +({re_f})',
                        dtype=np.float64
                    ),
                    Quantity(
                        'x_fleur_tot_nucl_charge',
                        rf'total nuclear charge += +({re_f})',
                        dtype=np.float64
                    ),
                ])
            ),
            Quantity(
                'scf_iteration',
                r'(it= +\d+[\s\S]+?it= +\d+ +is completed)',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity(
                        'energy_sum_eigenvalues',
                        rf'sum of eigenvalues = +({re_f})', dtype=np.float64, unit=ureg.hartree
                    ),
                    Quantity(
                        'energy_x_fleur_density_coulomb_potential',
                        rf'density\-coulomb potential int +({re_f}) htr', dtype=np.float64, unit=ureg.hartree
                    ),
                    Quantity(
                        'energy_x_fleur_density_effective_potential',
                        rf'density-effective potential i +({re_f}) htr', dtype=np.float64, unit=ureg.hartree
                    ),
                    Quantity(
                        'energy_x_fleur_charge_density_xc',
                        rf'charge density\-ex\.\-corr\.energ +({re_f}) htr', dtype=np.float64, unit=ureg.hartree
                    ),
                    Quantity(
                        'energy_total',
                        rf'total energy= +({re_f}) htr', dtype=np.float64, unit=ureg.hartree
                    ),
                    Quantity(
                        'energy_free',
                        rf'free energy= +({re_f}) htr', dtype=np.float64, unit=ureg.hartree
                    ),
                    Quantity(
                        'energy_total_t0',
                        rf'total electron energy= +({re_f}) htr', dtype=np.float64, unit=ureg.hartree
                    ),
                    Quantity(
                        'forces',
                        rf'FX_TOT= *({re_f}) FY_TOT= *({re_f}) FZ_TOT= *({re_f})',
                        repeats=True
                    ),
                    Quantity(
                        'kpoints',
                        rf'k=\( +({re_f} +{re_f} +{re_f})\):',
                        repeats=True, dtype=np.dtype(np.float64)
                    ),
                    Quantity(
                        'eigenvalues',
                        r'the *\d+ eigenvalues are:\s+([\s\d\.\-]+)',
                        repeats=True, dtype=np.dtype(np.float64)
                    ),
                    Quantity(
                        'fermi',
                        rf'new fermi energy +: +({re_f}) htr',
                        unit=ureg.hartree, dtype=np.float64
                    ),
                    Quantity(
                        'x_fleur_n_occupied_states',
                        r'number of occ\. states +: +(\d+)',
                        dtype=np.int32
                    ),
                    Quantity(
                        'x_fleur_valence_charge',
                        rf'valence charge +: +({re_f})',
                        dtype=np.float64
                    ),
                ])
            )
        ]


class FleurParser:
    def __init__(self):
        self.out_parser = OutParser()
        self._xc_map = {
            'pbe': ['GGA_X_PBE', 'GGA_C_PBE'],
            'rpbe': ['GGA_X_PBE', 'GGA_C_PBE'],
            'Rpbe': ['GGA_X_RPBE'],
            'pw91': ['GGA_X_PW91', 'GGA_C_PW91'],
            'l91': ['LDA_C_PW', 'LDA_C_PW_RPA', 'LDA_C_PW_MOD', 'LDA_C_OB_PW'],
            'vwn': ['LDA_C_VWN', 'LDA_C_VWN_1', 'LDA_C_VWN_2', 'LDA_C_VWN_3', 'LDA_C_VWN_4', 'LDA_C_VWN_RPA'],
            'bh': ['LDA_C_VBH'],
            'pz': ['LDA_C_PZ'],
            'x-a': [],
            'mjw': [],
            'wign': [],
            'hl': [],
            'xal': [],
            'relativistic': []
        }
        # TODO implement xml parser

    def init_parser(self):
        self.out_parser.mainfile = self.filepath
        self.out_parser.logger = self.logger

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.logger = logging.getLogger(__name__) if logger is None else logger
        self.maindir = os.path.dirname(self.filepath)
        self.init_parser()

        header = self.out_parser.get('header', {})
        sec_run = self.archive.m_create(Run)
        sec_run.program = Program(name='fleur', version=header.get('program_version'))
        sec_header = sec_run.m_create(x_fleur_header)
        for key, val in header.items():
            if key.startswith('x_fleur'):
                setattr(sec_header, key, val)

        sec_method = sec_run.m_create(Method)
        sec_method.basis_set.append(BasisSet(type='(L)APW+lo'))
        input = self.out_parser.get('input')
        if input is not None:
            for key in ['parameters', 'input_parameters']:
                setattr(sec_method, f'x_fleur_{key}', {
                    key: list([float(v) for v in val]) if isinstance(
                        val, np.ndarray) else val for key, val in input.get(key, {}).get('key_val', [])})

        electronic = self.out_parser.electronic
        if electronic is not None:
            sec_method.x_fleur_eigenvalues_parameters = {
                key: val for key, val in electronic.get('eigenvalues_parameters', {}).get('key_val')}
            sec_method.electronic = Electronic(smearing=Smearing(
                kind=electronic.smearing, width=electronic.width))

        exchange_correlation = self.out_parser.get('system', {}).get('exchange_correlation')
        if exchange_correlation is not None:
            exchange_correlation = [xc.strip() for xc in exchange_correlation.strip().split(' ', 1)]
            sec_method.dft = DFT()
            sec_xc_functional = sec_method.dft.m_create(XCFunctional)
            for xc_functional in self._xc_map.get(exchange_correlation[0], []):
                if '_X_' in xc_functional or xc_functional.endswith('_X'):
                    sec_xc_functional.exchange.append(Functional(name=xc_functional))
                elif '_C_' in xc_functional or xc_functional.endswith('_C'):
                    sec_xc_functional.correlation.append(Functional(name=xc_functional))
                elif 'HYB' in xc_functional:
                    sec_xc_functional.hybrid.append(Functional(name=xc_functional))
                else:
                    sec_xc_functional.contributions.append(Functional(name=xc_functional))
            sec_xc_functional.x_fleur_xc_correction = exchange_correlation[1]

        electric_field = self.out_parser.electic_field_parameters
        if electric_field is not None:
            for key, val in electric_field.items():
                setattr(sec_method, key, val)

        system = self.out_parser.system
        if system is not None:
            sec_system = sec_run.m_create(System)
            sec_system.atoms = Atoms(
                lattice_vectors=system.cell,
                labels=system.get('atoms', [None, None])[0],
                positions=system.get('atoms', [None, None])[1])

            sec_system.x_fleur_parameters = {key: val for key, val in system.get('parameters', {}).get('key_val', [])}

            for key, val in system.items():
                if key.startswith('x_fleur'):
                    setattr(sec_system, key, val)

        sec_scc = sec_run.m_create(Calculation)
        sec_scc.system_ref = sec_system
        scf = None
        for scf in self.out_parser.get('scf_iteration', []):
            sec_scf = sec_scc.m_create(ScfIteration)
            sec_scf_energy = sec_scf.m_create(Energy)
            sec_eigenvalues = sec_scf.m_create(BandEnergies)
            nspin = self.out_parser.get('system', {}).get('parameters', {}).get('jspins', 1)
            for key, val in scf.items():
                if val is None:
                    continue
                if key.startswith('energy_'):
                    setattr(sec_scf_energy, key.replace('energy_', ''), EnergyEntry(value=val))
                elif key == 'forces':
                    sec_scf.forces = Forces(total=ForcesEntry(value=val * ureg.hartree / ureg.bohr))
                elif key == 'fermi':
                    sec_scf_energy.fermi = val
                elif key == 'kpoints':
                    sec_eigenvalues.kpoints = val
                elif key == 'eigenvalues':
                    sec_eigenvalues.energies = np.reshape(val, (nspin, len(val) // nspin, len(val[0]))) * ureg.hartree
                elif key.startswith('x_fleur'):
                    setattr(sec_scf, key, val)

        if scf is not None:
            sec_scc.energy = sec_scf.energy
            sec_scc.forces = sec_scf.forces
            sec_scc.eigenvalues.append(sec_scf.eigenvalues[0])



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from nptyping import NDArray
import numpy as np
import ase

from nomad.datamodel.metainfo.simulation.calculation import (
    BandStructure, BandGap
)
from nomad.datamodel.metainfo.simulation.system import System
from nomad.normalizing.normalizer import Normalizer
from nomad import config, atomutils
from nomad.constants import pi


class BandStructureNormalizer(Normalizer):
    """Normalizer with the following responsibilities:

      - Calculates band gaps and energy references.
      - TODO: Creates labels for special points within the band path (band_segm_labels)
      - TODO: Determines if the path is a standard one or not (is_standard)
    """
    def normalize(self, logger=None) -> None:
        # Setup logger
        if logger is not None:
            self.logger = logger.bind(normalizer=self.__class__.__name__)

        # Do nothing if section run is not present
        if self.section_run is None:
            return

        # Loop through the bands
        for scc in self.section_run.calculation:

            # In order to resolve band gaps, we need a reference to the highest
            # occupied energy or the Fermi energy
            energy_fermi = scc.energy.fermi if scc.energy is not None else None
            energy_highest = scc.energy.highest_occupied if scc.energy is not None else None
            energy_lowest = scc.energy.lowest_unoccupied if scc.energy is not None else None

            # In order to resolve the special points and the reciprocal cell,
            # we need information about the system.
            system = scc.system_ref
            for band in scc.band_structure_electronic:
                valid_band = self.validate_band(band)
                if valid_band:
                    self.add_reciprocal_cell(band, system)
                    self.add_band_gap(
                        band,
                        energy_fermi,
                        energy_highest,
                        energy_lowest
                    )
                    self.add_path_labels(band, system)

    def validate_band(self, band: BandStructure) -> bool:
        """Used to check that a band has all required information for normalization.
        """
        if len(band.segment) == 0:
            self.logger.info("Could not normalize band structure as band segments are missing.")
            return False
        for segment in band.segment:
            seg_k_points = segment.kpoints
            seg_energies = segment.energies
            if seg_k_points is None or seg_energies is None:
                self.logger.info("Could not normalize band structure as energies or k points are missing.")
                return False
        return True

    def add_reciprocal_cell(self, band: BandStructure, system: System):
        """A reciprocal cell for this calculation. If the original unit cell is
        not a primitive one, then we will use the one given by spglib.

        If the used code is calculating it's own primitive cell, then the
        reciprocal cell used in e.g. band structure calculation might differ
        from the one given by spglib. To overcome this we would need to get the
        exact reciprocal cell used by the calculation or then test for
        different primitive cells
        (https://atztogo.github.io/spglib/definition.html#transformation-to-the-primitive-cell)
        whether the k-point path stays inside the first Brillouin zone.
        """
        # If reciprocal cell is reported by parser, use it.
        if band.reciprocal_cell is not None:
            return
        try:
            orig_atoms = system.m_cache["representative_atoms"]
            symmetry_analyzer = system.symmetry[0].m_cache["symmetry_analyzer"]
            prim_atoms = symmetry_analyzer.get_primitive_system()
        except Exception:
            self.logger.info("Could not resolve reciprocal cell.")
            return

        primitive_cell = prim_atoms.get_cell()
        source_cell = orig_atoms.get_cell()

        volume_primitive = primitive_cell.volume
        volume_source = source_cell.volume
        volume_diff = abs(volume_primitive - volume_source)

        if volume_diff > (0.001)**3:
            recip_cell = primitive_cell.reciprocal() * 1e10
        else:
            recip_cell = source_cell.reciprocal() * 1e10

        band.reciprocal_cell = recip_cell

    def get_k_space_distance(self, reciprocal_cell: NDArray, point1: NDArray, point2: NDArray) -> float:
        """Used to calculate the Euclidean distance of two points in k-space,
        given relative positions in the reciprocal cell.

        Args:
            reciprocal_cell: Reciprocal cell.
            point1: The first position in k-space.
            point2: The second position in k-space.

        Returns:
            float: Euclidian distance of the two points in k-space in SI units.
        """
        k_point_displacement = np.dot(reciprocal_cell, point1 - point2)
        k_point_distance = np.linalg.norm(k_point_displacement)

        return k_point_distance

    def add_band_gap(
            self,
            band: BandStructure,
            energy_fermi: NDArray,
            energy_highest: NDArray,
            energy_lowest: NDArray) -> None:
        """Given the band structure and information about energy references,
        determines the band gap and energy references separately for all spin
        channels.
        """
        band.energy_fermi = energy_fermi

        # No reference data available
        eref = energy_highest if energy_fermi is None else energy_fermi
        if eref is None:
            self.logger.info("could not resolve energy references or band gaps for band structure")
            return
        eref = eref.magnitude

        # Create energy reference sections for each spin channel, add fermi
        # energy if present
        n_channels = band.segment[0].energies.shape[0]
        for i_channel in range(n_channels):
            info = band.band_gap[i_channel] if len(band.band_gap) > i_channel else band.m_create(BandGap)
            info.index = i_channel
            if energy_highest is not None:
                info.energy_highest_occupied = energy_highest
            if energy_lowest is not None:
                info.energy_lowest_unoccupied = energy_lowest

        # Use a reference energy (fermi or highest occupied) to determine the
        # energy references from the band structure (discretization will affect
        # the exact location).
        reciprocal_cell = band.reciprocal_cell
        if reciprocal_cell is None:
            self.logger.info("could not resolve band gaps as reciprocal cell is missing")
            return

        # Gather the energies and k points from each segment into one big
        # array
        reciprocal_cell = reciprocal_cell.magnitude
        path: NDArray = []
        energies: NDArray = []
        for segment in band.segment:
            seg_k_points = segment.kpoints
            seg_energies = segment.energies
            seg_energies = seg_energies.magnitude
            seg_energies = np.swapaxes(seg_energies, 1, 2)
            path.append(seg_k_points)
            energies.append(seg_energies)

        path = np.concatenate(path, axis=0)
        energies = np.concatenate(energies, axis=2)

        # Handle spin channels separately to find gaps for spin up and down
        for i_channel in range(n_channels):
            i_energy_highest = None
            i_energy_lowest = None
            channel_energies = energies[i_channel, :, :]
            num_bands = channel_energies.shape[0]
            band_indices = np.arange(num_bands)
            band_minima_idx = channel_energies.argmin(axis=1)
            band_maxima_idx = channel_energies.argmax(axis=1)
            band_minima = channel_energies[band_indices, band_minima_idx]
            band_maxima = channel_energies[band_indices, band_maxima_idx]

            # Add a tolerance to minima and maxima
            band_minima_tol = band_minima + config.normalize.band_structure_energy_tolerance
            band_maxima_tol = band_maxima - config.normalize.band_structure_energy_tolerance

            for band_idx in range(num_bands):
                band_min = band_minima[band_idx]
                band_max = band_maxima[band_idx]
                band_min_tol = band_minima_tol[band_idx]
                band_max_tol = band_maxima_tol[band_idx]

                # If any of the bands band crosses the Fermi level, there is no
                # band gap
                if band_min_tol <= eref and band_max_tol >= eref:
                    break
                # Whole band below Fermi level, save the current highest
                # occupied band point
                elif band_min_tol <= eref and band_max_tol <= eref:
                    i_energy_highest = band_max
                    gap_lower_idx = band_maxima_idx[band_idx]
                # Whole band above Fermi level, save the current lowest
                # unoccupied band point
                elif band_min_tol >= eref:
                    i_energy_lowest = band_min
                    gap_upper_idx = band_minima_idx[band_idx]
                    break

            # Save the found energy references
            if i_energy_highest is not None:
                band.band_gap[i_channel].energy_highest_occupied = i_energy_highest
            if i_energy_lowest is not None:
                band.band_gap[i_channel].energy_lowest_unoccupied = i_energy_lowest

            # If highest occupied energy and a lowest unoccupied energy are
            # found, and the difference between them is positive, save
            # information about the band gap.
            gap_value = 0.0
            info = band.band_gap[i_channel]
            if i_energy_lowest is not None and i_energy_highest is not None:
                gap_value = float(i_energy_lowest - i_energy_highest)
                if gap_value > 0:
                    # See if the gap is direct or indirect by comparing the k-point
                    # locations with some tolerance
                    k_point_lower = path[gap_lower_idx]
                    k_point_upper = path[gap_upper_idx]
                    k_point_distance = self.get_k_space_distance(reciprocal_cell, k_point_lower, k_point_upper)
                    is_direct_gap = k_point_distance <= config.normalize.k_space_precision

                    band_gap_type = "direct" if is_direct_gap else "indirect"
                    info.type = band_gap_type
            info.value = gap_value

    def add_path_labels(self, band: BandStructure, system: System) -> None:
        """Adds special high symmmetry point labels to the band path. Only k
        points that land on the special points defined by Setyawan/Curtarolo
        are automatically labeled.
        """
        # If labels are already set by the parser dot nothing.
        for segment in band.segment:
            labels = segment.endpoints_labels
            if labels is not None:
                self.logger.info("Existing band segment labels detected, skipping label detection.")
                return

        # Try to get the required data. Fail if not found.
        try:
            cell = system.atoms.lattice_vectors.to("angstrom").magnitude
            reciprocal_cell_trans = band.reciprocal_cell.magnitude.T
            bravais_lattice = system.symmetry[0].bravais_lattice
        except Exception:
            self.logger.info("Could not resolve path labels as required information is missing.")
            return

        # Find special points for this lattice. If an error occurs, the labels
        # are simply not written.
        try:
            special_points = self.get_special_points(bravais_lattice, cell)
        except Exception as e:
            self.logger.warning("Could not resolve high-symmetry points for the given simulation cell.", exception=e)
            return

        # Form a contiguous array of k points for faster operations
        special_point_labels = list(special_points.keys())
        special_k_points = np.empty((len(special_points), 3))
        for i, kpt in enumerate(special_points.values()):
            special_k_points[i, :] = kpt
        special_k_points_cartesian = np.dot(special_k_points, reciprocal_cell_trans)

        # Match tolerance in 1/m. Taken from the VASP parser.
        eps = config.normalize.k_space_precision

        # Try to find matches for the special points. We only attempt to match
        # points at the start and end of a segment. Any labels set by the
        # parser are overridden, because one cannot ascertain that those labels
        # are consistent across codes.
        for segment in band.segment:
            start_point_cartesian = np.dot(segment.kpoints[0], reciprocal_cell_trans)
            end_point_cartesian = np.dot(segment.kpoints[-1], reciprocal_cell_trans)

            # Calculate distance in cartesian space
            start_index = atomutils.find_match(start_point_cartesian, special_k_points_cartesian, eps)
            end_index = atomutils.find_match(end_point_cartesian, special_k_points_cartesian, eps)

            if start_index is None:
                start_label = ""
            else:
                start_label = special_point_labels[start_index]
            if end_index is None:
                end_label = ""
            else:
                end_label = special_point_labels[end_index]
            segment.endpoints_labels = [start_label, end_label]

    def get_special_points(self, bravais_lattice, cell, eps=1e-4):
        """Return dict of special points.

        The definitions are from a paper by Wahyu Setyawana and Stefano
        Curtarolo:

            http://dx.doi.org/10.1016/j.commatsci.2010.05.010

        bravais_lattice: str
            bravais lattice in Pearson notation.
        cell: 3x3 ndarray
            Unit cell.
        eps: float
            Tolerance for cell-check.
        """
        # Special points that do not depend on lattice parameters. TODO: A lot
        # of the bravais lattice are missing from this implementation that is
        # copied from the VASP parser.
        special_points = {
            'cP': {
                'Γ': [0, 0, 0],
                'M': [1 / 2, 1 / 2, 0],
                'R': [1 / 2, 1 / 2, 1 / 2],
                'X': [0, 1 / 2, 0]
            },
            'cF': {
                'Γ': [0, 0, 0],
                'K': [3 / 8, 3 / 8, 3 / 4],
                'L': [1 / 2, 1 / 2, 1 / 2],
                'U': [5 / 8, 1 / 4, 5 / 8],
                'W': [1 / 2, 1 / 4, 3 / 4],
                'X': [1 / 2, 0, 1 / 2]
            },
            'cI': {
                'Γ': [0, 0, 0],
                'H': [1 / 2, -1 / 2, 1 / 2],
                'P': [1 / 4, 1 / 4, 1 / 4],
                'N': [0, 0, 1 / 2]
            },
            'tP': {
                'Γ': [0, 0, 0],
                'A': [1 / 2, 1 / 2, 1 / 2],
                'M': [1 / 2, 1 / 2, 0],
                'R': [0, 1 / 2, 1 / 2],
                'X': [0, 1 / 2, 0],
                'Z': [0, 0, 1 / 2]
            },
            'oP': {
                'Γ': [0, 0, 0],
                'R': [1 / 2, 1 / 2, 1 / 2],
                'S': [1 / 2, 1 / 2, 0],
                'T': [0, 1 / 2, 1 / 2],
                'U': [1 / 2, 0, 1 / 2],
                'X': [1 / 2, 0, 0],
                'Y': [0, 1 / 2, 0],
                'Z': [0, 0, 1 / 2]
            },
            'hP': {
                'Γ': [0, 0, 0],
                'A': [0, 0, 1 / 2],
                'H': [1 / 3, 1 / 3, 1 / 2],
                'K': [1 / 3, 1 / 3, 0],
                'L': [1 / 2, 0, 1 / 2],
                'M': [1 / 2, 0, 0]
            }
        }

        cellpar = ase.geometry.cell_to_cellpar(cell=cell)
        abc = cellpar[:3]
        angles = cellpar[3:] / 180 * pi
        a, b, c = abc
        alpha, _, gamma = angles

        # Check that the unit cells are as in the Setyawana-Curtarolo paper:
        if bravais_lattice == 'cP':
            assert abc.ptp() < eps and abs(angles - pi / 2).max() < eps
        elif bravais_lattice == 'cF':
            assert abc.ptp() < eps and abs(angles - pi / 3).max() < eps
        elif bravais_lattice == 'cI':
            angle = np.arccos(-1 / 3)
            assert abc.ptp() < eps and abs(angles - angle).max() < eps
        elif bravais_lattice == 'tP':
            assert abs(a - b) < eps and abs(angles - pi / 2).max() < eps
        elif bravais_lattice == 'oP':
            assert abs(angles - pi / 2).max() < eps
        elif bravais_lattice == 'hP':
            assert abs(a - b) < eps
            assert abs(gamma - pi / 3 * 2) < eps
            assert abs(angles[:2] - pi / 2).max() < eps
        elif bravais_lattice == 'mP':
            sin_alpha = np.sin(alpha)
            cos_alpha = np.cos(alpha)
            assert c >= a and c >= b
            assert alpha < pi / 2
            assert alpha < pi / 2
            assert (np.abs(angles[1:] - pi / 2) < eps).all()
            eta = (1 - b * cos_alpha / c) / (2 * sin_alpha**2)
            nu = 1 / 2 - eta * c * cos_alpha / b

            return {
                'Γ': [0, 0, 0],
                'A': [1 / 2, 1 / 2, 0],
                'C': [0, 1 / 2, 1 / 2],
                'D': [1 / 2, 0, 1 / 2],
                'D1': [1 / 2, 0, -1 / 2],
                'E': [1 / 2, 1 / 2, 1 / 2],
                'H': [0, eta, 1 - nu],
                'H1': [0, 1 - eta, nu],
                'H2': [0, eta, -nu],
                'M': [1 / 2, eta, 1 - nu],
                'M1': [1 / 2, 1 - eta, nu],
                'M2': [1 / 2, eta, -nu],
                'X': [0, 1 / 2, 0],
                'Y': [0, 0, 1 / 2],
                'Y1': [0, 0, -1 / 2],
                'Z': [1 / 2, 0, 0]
            }
        return special_points[bravais_lattice]



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import asyncio
from asyncio import Semaphore
import time
from typing import Any, Dict, List, Tuple

import httpx
from httpx import Timeout
from keycloak import KeycloakOpenID

from nomad import config, metainfo as mi
from nomad.datamodel import EntryArchive


def _collect(required, parent_section: mi.Section, parent_path: str = None) -> set:
    '''
    Flatten required quantities for uncoupled query
    '''

    quantities: set = set()

    if not isinstance(required, dict):
        return quantities

    for key, value in required.items():
        # some keys may index, get the exact name
        definition_name = key.split('[')[0]

        # validate definition name
        definition = parent_section.all_properties.get(definition_name)
        if definition is None:
            raise KeyError(f'{definition_name} is not a property of {parent_section}.')

        if parent_path:
            definition_name = f'{parent_path}.{definition_name}'

        quantities.add(definition_name)

        if isinstance(definition, mi.SubSection):
            quantities.update(_collect(value, definition.section_def, definition_name))
        elif isinstance(definition, mi.Quantity) and isinstance(definition.type, mi.Reference):
            next_parent_section = definition.type.target_section_def.m_resolved()
            parent_path = next_parent_section.path
            if parent_path in ['__ambiguous__', '__no_archive_path__']:
                continue
            quantities.update(_collect(value, next_parent_section, parent_path))

    return quantities


class ArchiveQuery:
    '''
    The async implementation works well with a large number of uploads, each has a small number of
    entries.

    Authentication is created by using valid `username` and `password`.
    If any is invalid, authenticated access is not available.

    Params:
        owner (str): ownership scope
        query (dict): query
        required (dict): required properties
        url (str): server url, if not specified, the default official NOMAD one is used
        after (str): specify the starting upload id to query, if users have knowledge of uploads,
            they may wish to start from a specific upload, default: ''
        results_max (int): maximum results to query, default: 1000
        page_size (int): size of page in each query, default: 100
        username (str): username for authenticated access, default: ''
        password (str): password for authenticated access, default: ''
        retry (int): number of retry when fetching uploads, default: 4
        sleep_time (float): sleep time for retry, default: 1.
    '''

    def __init__(
            self, owner: str = 'visible', query: dict = None, required: dict = None,
            url: str = None, after: str = None, results_max: int = 1000, page_size: int = 10,
            username: str = None, password: str = None, retry: int = 4, sleep_time: float = 4,
            from_api: bool = False):
        self._owner: str = owner
        self._required = required if required else dict(run='*')
        self._query_list: List[dict] = []
        if query:
            self._query_list.append(query)
        self._query_list.append({'quantities': list(_collect(self._required, EntryArchive.m_def))})
        self._url: str = url if url else config.client.url
        self._after: str = after
        self._results_max: int = results_max if results_max > 0 else 1000
        self._page_size: int = page_size if page_size > 0 else 10
        self._retry: int = retry if retry >= 0 else 4
        self._sleep_time: float = sleep_time if sleep_time > 0. else 1.

        self._username: str = username
        self._password: str = password
        self._from_api: bool = from_api

        self._oidc = KeycloakOpenID(
            server_url=config.keycloak.public_server_url, realm_name=config.keycloak.realm_name,
            client_id=config.keycloak.client_id)

        # noinspection PyTypeChecker
        self._token: dict = None

        # local data storage
        self._uploads: List[Tuple[str, int]] = []
        self._current_after: str = self._after
        self._current_results: int = 0

    @property
    def _query(self) -> dict:
        return {'and': self._query_list}

    @property
    def _fetch_url(self) -> str:
        return f'{self._url}/v1/entries/query'

    @property
    def _download_url(self) -> str:
        return f'{self._url}/v1/entries/archive/query'

    @property
    def _auth_url(self) -> str:
        return f'{self._url}/v1/auth/token'

    def _update_token_from_api(self):
        if self._token:
            return

        response = httpx.get(
            self._auth_url, params=dict(username=self._username, password=self._password))

        if response.status_code != 200:
            response_json = response.json()
            reason = response_json.get("description") or response_json.get(
                "detail") or "unknown reason"
            raise ValueError(f'Could not authenticate: {reason} ({response.status_code})')

        self._token = response.json()

    # noinspection DuplicatedCode
    def _update_token_from_keycloak(self):
        if self._token is None:
            self._token = self._oidc.token(username=self._username, password=self._password)
            self._token['time'] = time.time()
        elif self._token['expires_in'] + self._token['time'] < int(time.time()) + 10:
            # noinspection PyBroadException
            try:
                self._token = self._oidc.refresh_token(self._token['refresh_token'])
                self._token['time'] = time.time()
            except Exception:
                self._token = self._oidc.token(username=self._username, password=self._password)
                self._token['time'] = time.time()

    def _auth(self) -> dict:
        if (self._username is None) or (self._password is None):
            return {}

        if self._from_api:
            self._update_token_from_api()
        else:
            self._update_token_from_keycloak()

        return dict(Authorization=f'Bearer {self._token["access_token"]}')

    @property
    def _fetch_request(self) -> dict:
        '''
        Generate fetch request.
        '''

        request = {
            'owner': self._owner,
            'query': self._query,
            'pagination': {
                'page_size': 0
            },
            'aggregations': {
                'uploads': {
                    'terms': {
                        'quantity': 'upload_id',
                        'pagination': {
                            'page_size': self._page_size,
                            'page_after_value': self._current_after
                        }
                    }
                }
            }
        }

        # print(f'Current request: {request}')

        return request

    def _download_request(self, upload_id: str, upload_count: int) -> dict:
        '''
        Generate download request.
        '''

        request: Dict[str, Any] = dict(owner=self._owner, required=self._required)
        request['query'] = {'and': []}
        for t_list in self._query_list:
            request['query']['and'].append(t_list)
        request['query']['and'].append({'upload_id': upload_id})
        request.setdefault('pagination', {'page_size': upload_count})

        # print(f'Current request: {request}')

        return request

    def clear(self):
        '''
        Clear all fetched and downloaded data. Users can then call .fetch() and .download() again.
        '''

        self._uploads: List[Tuple[str, Any]] = []  # (id, count)
        self._current_after = self._after
        self._current_results = 0

    async def _fetch_async(self, number: int = 0) -> int:
        '''
        There is no need to perform fetching asynchronously as the required number of uploads
        depends on previous queries.

        It is just a wrapper to avoid the `requests` library.

        Params:
            number (int): approx. number of **entries** to be fetched

        Returns:
            The number of entries fetched
        '''

        # if maximum number of entries have been previously fetched
        # not going to fetch more entries
        if self._current_results >= self._results_max:
            return 0

        # get all entries at once
        if number == 0:
            number = self._results_max

        num_retry: int = 0
        num_entry: int = 0

        async with httpx.AsyncClient(timeout=Timeout(timeout=300)) as session:
            while True:
                response = await session.post(
                    self._fetch_url, json=self._fetch_request, headers=self._auth())

                if response.status_code >= 400:
                    if response.status_code < 500:
                        response_json = response.json()
                        reason = response_json.get("description") or response_json.get(
                            "detail") or "unknown reason"
                        raise ValueError(f'Server returns {response.status_code}: {reason}')
                    if response.status_code in [500, 502, 504]:
                        if num_retry > self._retry:
                            print('Maximum retry reached.')
                            break
                        else:
                            print(f'Retrying in {self._sleep_time} seconds...')
                            time.sleep(self._sleep_time)
                            continue

                response_json = response.json()

                data = response_json['aggregations']['uploads']['terms']
                header = [(bucket['value'], int(bucket['count'])) for bucket in data['data']]
                self._current_after = data['pagination'].get('next_page_after_value', None)

                current_size: int = sum([count for _, count in header])

                # no more entries
                if current_size == 0:
                    break

                if self._current_results + current_size >= self._results_max:
                    # current query has sufficient entries to exceed the limit
                    for upload_id, count in header:
                        self._current_results += count
                        num_entry += count
                        # required number of entries have been acquired
                        if self._current_results >= self._results_max:
                            entry_difference = self._current_results - self._results_max
                            self._current_results -= entry_difference
                            num_entry -= entry_difference
                            self._uploads.append((upload_id, count - entry_difference))
                            self._current_after = upload_id
                            break
                        else:
                            self._uploads.append((upload_id, count))
                    break
                else:
                    # current query should be added
                    num_entry += current_size
                    self._current_results += current_size
                    self._uploads.extend(header)

                    # if exceeds the required number, exit
                    # `self._current_after` is automatically set
                    if num_entry >= number:
                        break

        print(f'{num_entry} entries are qualified and added to the download list.')

        return num_entry

    async def _download_async(self, number: int = 0) -> List[EntryArchive]:
        '''
        Download required entries asynchronously.

        Params:
            number (int): number of **entries** to download

        Returns:
            A list of EntryArchive
        '''

        num_entry: int = 0
        num_upload: int = 0
        for _, count in self._uploads:
            num_entry += count
            num_upload += 1
            if num_entry >= number:
                break

        semaphore = Semaphore(30)

        async with httpx.AsyncClient(timeout=Timeout(timeout=300)) as session:
            tasks = [asyncio.create_task(
                self._acquire(
                    upload, session, semaphore)) for upload in self._uploads[:num_upload]]
            results = await asyncio.gather(*tasks)

        # flatten 2D list
        return [result for sub_results in results for result in sub_results]

    async def _acquire(self, upload: Tuple[str, int], session, semaphore) -> List[EntryArchive]:
        '''
        Perform the download task.

        Params:
            upload (Tuple[str, int]): upload

            session (httpx.AsyncClient): httpx client

            semaphore (asyncio.Semaphore): semaphore

        Returns:
            A list of EntryArchive
        '''

        request = self._download_request(upload[0], upload[1])

        async with semaphore:
            response = await session.post(self._download_url, json=request, headers=self._auth())
            if response.status_code >= 400:
                print(
                    f'Request with upload id {upload[0]} returns {response.status_code},'
                    f' will retry in the next download call...')
                return []

            # successfully downloaded data
            # extract and remove the corresponding id in the list
            response_json = response.json()

            self._uploads.remove(upload)

            result = [EntryArchive.m_from_dict(
                result['archive']) for result in response_json['data']]

            if not result:
                print(f'No result returned for id {upload[0]}, is the query proper?')

            return result

    def fetch(self, number: int = 0) -> int:
        '''
        Fetch uploads from remote.

        Params:
            number (int): number of **entries** to fetch

        Returns:
            The number of entries fetched
        '''

        print('Fetching remote uploads...')

        return asyncio.run(self._fetch_async(number))

    def download(self, number: int = 0) -> List[EntryArchive]:
        '''
        Download fetched entries from remote.
        Automatically call .fetch() if not fetched.

        Params:
            number (int): number of **entries** to download at a single time

        Returns:
            A list of downloaded EntryArchive
        '''

        pending_size: int = sum([count for _, count in self._uploads])

        # download all at once
        if number == 0:
            number = pending_size

        if number == 0:
            # empty list, fetch as many as possible
            number = self.fetch()
        elif pending_size < number:
            # if not sufficient fetched entries, fetch first
            self.fetch(number - pending_size)

        print('Downloading required data...')

        return asyncio.run(self._download_async(number))

    async def async_fetch(self, number: int = 0) -> int:
        '''
        Asynchronous interface for use in a running event loop.
        '''

        print('Fetching remote uploads...')

        return await self._fetch_async(number)

    async def async_download(self, number: int = 0) -> List[EntryArchive]:
        '''
        Asynchronous interface for use in a running event loop.
        '''

        pending_size: int = sum([count for _, count in self._uploads])

        # download all at once
        if number == 0:
            number = pending_size

        if number == 0:
            # empty list, fetch as many as possible
            number = await self.async_fetch()
        elif pending_size < number:
            # if not sufficient fetched entries, fetch first
            await self.async_fetch(number - pending_size)

        print('Downloading required data...')

        return await self._download_async(number)

    def upload_list(self) -> List[Tuple[str, int]]:
        return self._uploads



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import typing

from nomad.metainfo import (
    MSection, Package, Quantity, SubSection)


m_package = Package(name='tabulartree')


class TabularTreeNodeInfo(MSection):
    value = Quantity(type=typing.Any)
    description = Quantity(type=str)
    unit = Quantity(type=str)


class TabularTreeLevel3(MSection):
    name = Quantity(type=str, default='<node name?>')
    info = SubSection(sub_section=TabularTreeNodeInfo)


class TabularTreeLevel2(MSection):
    name = Quantity(type=str, default='<node name?>')
    info = SubSection(sub_section=TabularTreeNodeInfo)
    nodes = SubSection(sub_section=TabularTreeLevel3, repeats=True)


class TabularTreeLevel1(MSection):
    name = Quantity(type=str, default='<node name?>')
    info = SubSection(sub_section=TabularTreeNodeInfo)
    nodes = SubSection(sub_section=TabularTreeLevel2, repeats=True)


class TabularTree(MSection):
    name = Quantity(type=str, default='<node name?>')
    info = SubSection(sub_section=TabularTreeNodeInfo)
    nodes = SubSection(sub_section=TabularTreeLevel1, repeats=True)


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference, MEnum, derived)
from nomad.datamodel.metainfo.simulation.method import Method
from nomad.datamodel.metainfo.simulation.system import System
from nomad.datamodel.metainfo.simulation.calculation import Calculation
from nomad.datamodel.metainfo.common import FastAccess

m_package = Package()


class AccessoryInfo(MCategory):
    '''
    Information that *in theory* should not affect the results of the calculations (e.g.,
    timing).
    '''

    m_def = Category()


class ProgramInfo(MCategory):
    '''
    Contains information on the program that generated the data, i.e. the program_name,
    program_version, program_compilation_host and program_compilation_datetime as direct
    children of this field.
    '''

    m_def = Category(categories=[AccessoryInfo])


class Program(MSection):
    '''
    Contains the specifications of the program.
    '''

    m_def = Section(validate=False)

    name = Quantity(
        type=str,
        shape=[],
        description='''
        Specifies the name of the program that generated the data.
        ''',
        categories=[AccessoryInfo, ProgramInfo])

    version = Quantity(
        type=str,
        shape=[],
        description='''
        Specifies the version of the program that was used. This should be the version
        number of an official release, the version tag or a commit id as well as the
        location of the repository.
        ''',
        categories=[AccessoryInfo, ProgramInfo])

    compilation_datetime = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        Contains the program compilation date and time from *Unix epoch* (00:00:00 UTC on
        1 January 1970) in seconds. For date and times without a timezone, the default
        timezone GMT is used.
        ''',
        categories=[AccessoryInfo, ProgramInfo])

    compilation_host = Quantity(
        type=str,
        shape=[],
        description='''
        Specifies the host on which the program was compiled.
        ''',
        categories=[AccessoryInfo, ProgramInfo])


class TimeRun(MSection):
    '''
    Contains information on timing information of the run.
    '''

    m_def = Section(validate=False)

    date_end = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        Stores the end date of the run as time since the *Unix epoch* (00:00:00 UTC on 1
        January 1970) in seconds. For date and times without a timezone, the default
        timezone GMT is used.
        ''')

    date_start = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        Stores the start date of the run as time since the *Unix epoch* (00:00:00 UTC on 1
        January 1970) in seconds. For date and times without a timezone, the default
        timezone GMT is used.
        ''')

    cpu1_end = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        Stores the end time of the run on CPU 1.
        ''')

    cpu1_start = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        Stores the start time of the run on CPU 1.
        ''')

    wall_end = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        Stores the internal wall-clock time at the end of the run.
        ''')

    wall_start = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        Stores the internal wall-clock time from the start of the run.
        ''')


class MessageRun(MSection):
    '''
    Contains warning, error, and info messages of the run.
    '''

    m_def = Section(validate=False)

    type = Quantity(
        type=str,
        shape=[],
        description='''
        Type of the message. Can be one of warning, error, info, debug.
        ''')

    value = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the message of the computational program, given by type.
        ''')


class Run(MSection):
    '''
    Every section run represents a single call of a program.
    '''

    m_def = Section(validate=False)

    calculation_file_uri = Quantity(
        type=str,
        shape=[],
        description='''
        Contains the nomad uri of a raw the data file connected to the current run. There
        should be an value for the main_file_uri and all ancillary files.
        ''')

    clean_end = Quantity(
        type=bool,
        shape=[],
        description='''
        Indicates whether this run terminated properly (true), or if it was killed or
        exited with an error code unequal to zero (false).
        ''')

    raw_id = Quantity(
        type=str,
        shape=[],
        description='''
        An optional calculation id, if one is found in the code input/output files.
        ''')

    starting_run_ref = Quantity(
        type=Reference(SectionProxy('Run')),
        shape=[],
        description='''
        Links the current section run to a section run containing the calculations from
        which the current section starts.
        ''',
        categories=[FastAccess])

    n_references = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
         Number of references to the current section calculation.
        ''')

    runs_ref = Quantity(
        type=Reference(SectionProxy('Run')),
        shape=['n_references'],
        description='''
        Links the the current section to other run sections. Such a link is necessary for
        example for workflows that may contain a series of runs.
        ''',
        categories=[FastAccess])

    program = SubSection(sub_section=Program.m_def)

    time_run = SubSection(sub_section=TimeRun.m_def)

    message = SubSection(sub_section=MessageRun.m_def)

    method = SubSection(sub_section=Method.m_def, repeats=True)

    system = SubSection(sub_section=System.m_def, repeats=True)

    calculation = SubSection(sub_section=Calculation.m_def, repeats=True)


m_package.__init_metainfo__()
